Package: rainette
Type: Package
Title: The Reinert Method for Textual Data Clustering
Version: 0.3.3
Date: 2026-05-04
Authors@R: c(person("Julien", "Barnier", email="julien.barnier@cnrs.fr",
    role=c("aut","cre")), person("Florian", "Privé",
    email="florian.prive.21@gmail.com", role="ctb"))
Maintainer: Julien Barnier <julien.barnier@cnrs.fr>
Description: An R implementation of the Reinert text clustering method. For more
    details about the algorithm see the included vignettes or Reinert (1990)
    <doi:10.1177/075910639002600103>.
License: GPL (>= 3)
VignetteBuilder: knitr
URL: https://juba.github.io/rainette/
BugReports: https://github.com/juba/rainette/issues
Encoding: UTF-8
Depends: R (>= 3.6.0)
Imports: dplyr (>= 1.1.0), tidyr, purrr, ggplot2, stringr, quanteda (>=
        2.1), quanteda.textstats, RSpectra, dendextend, ggwordcloud,
        gridExtra, rlang, shiny, miniUI, highr, progressr, Rcpp (>=
        1.0.3)
Suggests: testthat, knitr, rmarkdown, tm, FNN, vdiffr,
        quanteda.textmodels
RoxygenNote: 7.3.2
LinkingTo: Rcpp
NeedsCompilation: yes
Packaged: 2026-05-04 13:28:24 UTC; julien
Author: Julien Barnier [aut, cre],
  Florian Privé [ctb]
Repository: CRAN
Date/Publication: 2026-05-04 18:30:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-04 22:10:48 UTC; unix
Archs: rainette.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(split_segments,Corpus)
S3method(split_segments,character)
S3method(split_segments,corpus)
S3method(split_segments,tokens)
export(clusters_by_doc_table)
export(cutree)
export(cutree_rainette)
export(cutree_rainette2)
export(docs_by_cluster_table)
export(import_corpus_iramuteq)
export(merge_segments)
export(rainette)
export(rainette2)
export(rainette2_complete_groups)
export(rainette2_explor)
export(rainette2_plot)
export(rainette_explor)
export(rainette_plot)
export(rainette_stats)
export(split_segments)
import(dendextend)
import(dplyr)
import(ggplot2)
import(ggwordcloud)
import(miniUI)
import(quanteda)
import(shiny)
import(stringr)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(purrr,map_chr)
importFrom(purrr,map_int)
importFrom(rlang,.env)
importFrom(rlang,sym)
useDynLib(rainette)

# Rainette

[![CRAN status](https://www.r-pkg.org/badges/version-ago/rainette)](https://cran.r-project.org/package=rainette)
[![rainette status badge](https://juba.r-universe.dev/badges/rainette)](https://juba.r-universe.dev)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.6047058.svg)](https://doi.org/10.5281/zenodo.6047058)
![CRAN Downloads](https://cranlogs.r-pkg.org/badges/last-month/rainette)
[![R build status](https://github.com/juba/rainette/workflows/R-CMD-check/badge.svg)](https://github.com/juba/rainette/actions?query=workflow%3AR-CMD-check)

<!-- [![Coverage status](https://codecov.io/gh/juba/rainette/branch/master/graph/badge.svg)](https://codecov.io/github/juba/rainette?branch=master) -->

Rainette is an R package which implements a variant of the Reinert textual clustering method. This method is available in other softwares such as [Iramuteq](https://pratinaud.gitpages.huma-num.fr/iramuteq-website/) (free software) or [Alceste](https://www.image-zafar.com/Logiciel.html) (commercial, closed source).

## Features

- Simple and double clustering algorithms
- Plot functions and shiny interfaces to visualise and explore clustering results
- Utility functions to split a corpus into segments or import a corpus in Iramuteq format

## Installation

The package is installable from CRAN.

```r
install_packages("rainette")
```

The development version is installable from [R-universe](https://r-universe.dev).

```r
install.packages("rainette", repos = "https://juba.r-universe.dev")
```

## Usage

Let's start with an example corpus provided by the excellent [quanteda](https://quanteda.io) package.

```r
library(quanteda)
data_corpus_inaugural
```

First, we'll use `split_segments()` to split each document into segments of about 40 words (punctuation is taken into account).

```r
corpus <- split_segments(data_corpus_inaugural, segment_size = 40)
```

Next, we'll apply some preprocessing and compute a document-term matrix with `quanteda` functions.

```r
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 10)
```

We can then apply a simple clustering on this matrix with the `rainette()` function. We specify the number of clusters (`k`), and the minimum number of forms in each segment (`min_segment_size`). Segments which do not include enough forms will be merged with the following or previous one when possible.

```r
res <- rainette(dtm, k = 6, min_segment_size = 15)
```

We can use the `rainette_explor()` shiny interface to visualise and explore the different clusterings at each `k`.

```r
rainette_explor(res, dtm, corpus)
```

![rainette_explor() interface](man/figures/rainette_explor_plot_en.png)

The _Cluster documents_ tab allows to browse and filter the documents in each cluster.

![rainette_explor() documents tab](man/figures/rainette_explor_docs_en.png)

We can also directly generate the clusters description plot for a given `k` with `rainette_plot()`.

```r
rainette_plot(res, dtm, k = 5)
```

Or cut the tree at chosen `k` and add a group membership variable to our corpus metadata.

```r
docvars(corpus)$cluster <- cutree(res, k = 5)
```

In addition to this, we can also perform a double clustering, _ie_ two simple clusterings produced with different `min_segment_size` which are then "crossed" to generate more robust clusters. To do this, we use `rainette2()` on two `rainette()` results :

```r
res1 <- rainette(dtm, k = 5, min_segment_size = 10)
res2 <- rainette(dtm, k = 5, min_segment_size = 15)
res <- rainette2(res1, res2, max_k = 5)
```

We can then use `rainette2_explor()` to explore and visualise the results.

```r
rainette2_explor(res, dtm, corpus)
```

![rainette2_explor() interface](man/figures/rainette2_explor_en.png)

## Tell me more

Two vignettes are available :

- Introduction and usage vignette : [english](https://juba.github.io/rainette/articles/introduction_en.html), [french](https://juba.github.io/rainette/articles/introduction_usage.html)
- Algorithms description vignette : [english](https://juba.github.io/rainette/articles/algorithms_en.html), [french](https://juba.github.io/rainette/articles/algorithmes.html)

## Credits

This clustering method has been created by Max Reinert, and is described in several articles, notably :

- Reinert M., "Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte", _Cahiers de l'analyse des données_, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
- Reinert M., "Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval", _Bulletin de Méthodologie Sociologique_, Volume 26, Numéro 1, 1990. <https://doi.org/10.1177/075910639002600103>
- Reinert M., "Une méthode de classification des énoncés d’un corpus présentée à l’aide d’une application", _Les cahiers de l’analyse des données_, Tome 15, Numéro 1, 1990. <https://www.numdam.org/item/?id=CAD_1990__15_1_21_0>

Thanks to Pierre Ratineau, the author of [Iramuteq](https://pratinaud.gitpages.huma-num.fr/iramuteq-website/), for providing it as free software and open source. Even if the R code has been almost entirely rewritten, it has been a precious resource to understand the algorithms.

Many thanks to [Sébastien Rochette](https://github.com/statnmap) for the creation of the hex logo.

Many thanks to [Florian Privé](https://github.com/privefl/) for his work on rewriting and optimizing the Rcpp code.

library(quanteda)
context("plot functions")

# skip("Skipping plots")
skip_on_ci()
skip_on_cran()
skip_if_not_installed("vdiffr")
skip_if_not_installed("FNN")


mini_corpus <- head(data_corpus_inaugural, n = 10)
mini_corpus <- split_segments(mini_corpus)
tok <- tokens(mini_corpus, remove_punct = TRUE)
tok <- tokens_tolower(tok)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok)
dtm <- dfm_trim(dtm, min_termfreq = 3)

res1 <- rainette(dtm, k = 6, min_segment_size = 5, min_split_members = 3)
res2 <- rainette(dtm, k = 6, min_segment_size = 10, min_split_members = 3)
res12 <- rainette2(res1, res2, max_k = 6)

plot1 <- rainette_plot(res1, dtm, k = 6)
plot2 <- rainette_plot(res1, dtm, k = 5, free_scales = TRUE)
plot3 <- rainette_plot(res1, dtm, k = 5, measure = "lr")
plot3b <- rainette_plot(res1, dtm, k = 5, measure = "frequency")
plot3c <- rainette_plot(res1, dtm, k = 5, measure = "docprop")
plot4 <- rainette_plot(res1, dtm, k = 3, show_negative = FALSE)
plot5 <- rainette_plot(res1, dtm, k = 4, n_terms = 20, text_size = 9)

vdiffr::expect_doppelganger("Base rainette_plot", plot1)
vdiffr::expect_doppelganger("Base rainette_plot with free_scales", plot2)
vdiffr::expect_doppelganger("Base rainette_plot measure='lr'", plot3)
vdiffr::expect_doppelganger("Base rainette_plot measure='frequency'", plot3b)
vdiffr::expect_doppelganger("Base rainette_plot measure='docprop'", plot3c)
vdiffr::expect_doppelganger("Base rainette_plot with k and without negative", plot4)
vdiffr::expect_doppelganger("Base rainette_plot with k, n_terms and font_size", plot5)

plot8 <- rainette2_plot(res12, dtm, k = 5)
plot8b <- rainette2_plot(res12, dtm, k = 6)
plot9 <- rainette2_plot(res12, dtm, k = 4, free_scales = TRUE)
plot10 <- rainette2_plot(res12, dtm, k = 4, measure = "lr")
plot10b <- rainette2_plot(res12, dtm, k = 4, measure = "frequency")
plot10c <- rainette2_plot(res12, dtm, k = 4, measure = "docprop")
plot11 <- rainette2_plot(res12, dtm, k = 3, show_negative = FALSE)
plot12 <- rainette2_plot(res12, dtm, k = 4, n_terms = 20, text_size = 9)
plot13 <- rainette2_plot(res12, dtm, k = 4, complete_groups = TRUE)

vdiffr::expect_doppelganger("Base rainette2_plot", plot8)
vdiffr::expect_doppelganger("Base rainette2_plot with k > 5", plot8b)
vdiffr::expect_doppelganger("Base rainette2_plot with free_scales", plot9)
vdiffr::expect_doppelganger("Base rainette2_plot measure='lr'", plot10)
vdiffr::expect_doppelganger("Base rainette2_plot measure='frequency'", plot10b)
vdiffr::expect_doppelganger("Base rainette2_plot measure='docprop'", plot10c)
vdiffr::expect_doppelganger("Base rainette2_plot with k and without negative", plot11)
vdiffr::expect_doppelganger("Base rainette2_plot with k, n_terms and font_size", plot12)
vdiffr::expect_doppelganger("Base rainette2_plot with complete_groups", plot13)

## Extract texts from a list of documents in Iramuteq format
## `thematics` indicate if potential thematics have to been removed

extract_texts <- function(docs, thematics) {
  purrr::map_chr(docs, function(text) {
    ## Remove metadata
    text <- stringr::str_replace_all(text, "^\\*\\w+_.*?\n", "")
    ## Remove thematics
    if (thematics == "remove") {
      text <- stringr::str_replace_all(text, "(^|\n)-\\*[^ ].*?\n", "\\1")
    }
    text
  })
}

## Extract metadata from a list of documents in Iramuteq format
## and return them as a data.frame

extract_metadata <- function(docs) {
  meta <- purrr::map_dfr(docs, function(text) {
    first_line <- stringr::str_extract(text, "^.*?\n")
    vars_line <- stringr::str_extract_all(first_line, "\\*(.*?)[ \n$]")

    # No metadata for this doc
    if (identical(vars_line[[1]], NA_character_)) {
      return(data.frame(rainette_nometa = NA))
    }

    res <- purrr::map(vars_line, function(vars) {
      vars <- stringr::str_replace(vars, "^\\*", "")
      vars <- stringr::str_trim(vars)
      vars <- stringr::str_split_fixed(vars, "_", n = 2)
      values <- as.list(vars[, 2])
      names(values) <- vars[, 1]
      values
    })
    purrr::flatten_dfr(res)
  })
  meta[["rainette_nometa"]] <- NULL
  meta
}


#' Import a corpus in Iramuteq format
#'
#' @param f a file name or a connection
#' @param id_var name of metadata variable to be used as documents id
#' @param thematics if "remove", thematics lines are removed. If "split", texts as splitted at each thematic, and metadata duplicated accordingly
#' @param ... arguments passed to \code{\link[base:connections]{file}} if `f` is a file name.
#'
#' @return
#' A quanteda corpus object. Note that metadata variables in docvars are all imported as characters.
#'
#' @export
#' @import stringr
#' @importFrom rlang sym

import_corpus_iramuteq <- function(
  f,
  id_var = NULL,
  thematics = c("remove", "split"),
  ...
) {
  thematics <- match.arg(thematics)

  ## Open as a filename or a connection
  close_con <- FALSE
  if (is.character(f)) {
    close_con <- TRUE
    f <- file(f, ...)
  }
  if (!inherits(f, "connection")) {
    stop("f must be either a connection or a file path")
  }

  ## Read text
  lines <- readLines(f)
  if (close_con) {
    close(f)
  }
  lines <- lines[lines != ""]
  text <- paste(lines, collapse = "\n")

  ## Detect if corpus has metadata
  has_metadata <- stringr::str_detect(text, "(^|\n)\\*\\*\\*\\* +\\*\\w+_")
  ## Detect if corpus has thematics
  has_thematic <- stringr::str_detect(text, "(^|\n)-\\*[^ ]")

  ## Split into documents
  docs <- stringr::str_split(text, "(^|\n)\\*\\*\\*\\*(\n| +)", simplify = TRUE)
  docs <- docs[docs != ""]

  ## Extract texts
  texts <- extract_texts(docs, thematics)
  ## Extract metadata
  if (has_metadata) {
    metadata <- extract_metadata(docs)
  }

  ## Split by thematics
  if (has_thematic && thematics == "split") {
    thems <- purrr::map(
      stringr::str_match_all(docs, "(^|\n)-\\*(.+)\n"),
      ~ .x[, 3]
    )
    thems <- purrr::flatten_chr(thems)
    texts <- stringr::str_split(texts, "(^|\n)-\\*(.+)\n")
    texts <- purrr::imap_dfr(texts, function(text, i) {
      data.frame(id = i, text = text[text != ""], stringsAsFactors = FALSE)
    })
    ## Duplicate metadata and add thematics variable
    if (has_metadata) {
      metadata <- metadata[texts$id, ]
      metadata <- data.frame(
        metadata,
        thematics = thems,
        stringsAsFactors = FALSE
      )
      ## Generate id by thematic if necessary
      if (!is.null(id_var)) {
        metadata <- metadata %>%
          dplyr::group_by(.data[[id_var]]) %>%
          dplyr::mutate(
            rainette_split_id = paste(.data[[id_var]], 1:n(), sep = "_")
          )
      }
      id_var <- "rainette_split_id"
    } else {
      metadata <- data.frame(thematics = thems, stringsAsFactors = FALSE)
      has_metadata <- TRUE
    }
    texts <- texts$text
  }

  ## Build corpus
  corpus <- data.frame(text = texts, stringsAsFactors = FALSE)
  if (has_metadata) {
    corpus <- data.frame(metadata, corpus, check.names = FALSE)
  }
  if (!is.null(id_var)) {
    rownames(corpus) <- corpus[[id_var]]
    ## Remove computed id by thematic if present
    corpus[["rainette_split_id"]] <- NULL
  }

  quanteda::corpus(corpus, text_field = "text")
}

context("import Iramuteq corpus")

text <- "**** *id_01 *sexe_m *age_18 *prof_ouvrier
-*t1
Donec placerat quis quam ac interdum. Cras tempus eros et laoreet dictum. Morbi ac dapibus magna.
-*t2
Vivamus dictum, nisi et suscipit fermentum, tortor nulla tempor quam, a ullamcorper elit erat nec arcu.
**** *id_02 *sexe_f  *prof_agent_dentretien   *age_
-*t1
Vestibulum vitae ante eget justo maximus luctus. Nulla facilisi.

Integer tincidunt sed sapien mattis euismod. Proin sed massa neque.
-*t2
Nam nunc neque, fringilla vel pharetra sit amet, posuere a nibh.

**** *id_03 *sexe_f *age_26
-*t3
Maecenas semper dictum nisi.

Sed dolor urna, fringilla sed justo vel, fermentum dictum magna.

-*t4
Vestibulum mollis, ligula in semper vulputate, mauris diam aliquam est, ut placerat sem felis a nibh."

con <- textConnection(text)
corpus <- import_corpus_iramuteq(con, id_var = "id", thematics = "remove")
close(con)
con <- textConnection(text)
corpus_split_id <- import_corpus_iramuteq(con, id_var = "id", thematics = "split")
close(con)
con <- textConnection(text)
corpus_split <- import_corpus_iramuteq(con, thematics = "split")
close(con)

test_that("metadata imported correctly", {
  expect_equal(docvars(corpus, "id"), c("01", "02", "03"))
  expect_equal(docvars(corpus, "sexe"), c("m", "f", "f"))
  expect_equal(docvars(corpus, "age"), c("18", "", "26"))
  expect_equal(docvars(corpus, "prof"), c("ouvrier", "agent_dentretien", NA))
  expect_equal(docnames(corpus), docvars(corpus, "id"))
})

test_that("texts imported correctly", {
  expect_equal(ndoc(corpus), 3)
  expect_equal(as.character(corpus)[["02"]], "Vestibulum vitae ante eget justo maximus luctus. Nulla facilisi.\nInteger tincidunt sed sapien mattis euismod. Proin sed massa neque.\nNam nunc neque, fringilla vel pharetra sit amet, posuere a nibh.")
})


test_that("thematics splitted correctly", {
  expect_equal(ndoc(corpus_split), 6)
  expect_equal(ndoc(corpus_split_id), 6)
  expect_equal(as.character(corpus_split_id)[["02_2"]], "Nam nunc neque, fringilla vel pharetra sit amet, posuere a nibh.")
  expect_equal(docnames(corpus_split), paste0("text", 1:6))
  expect_equal(docnames(corpus_split_id), c("01_1", "01_2", "02_1", "02_2", "03_1", "03_2"))
  expect_equal(names(docvars(corpus_split)), names(docvars(corpus_split_id)))
  expect_equal(docvars(corpus_split, "age"), c("18", "18", "", "", "26", "26"))
  expect_equal(docvars(corpus_split, "age"), docvars(corpus_split_id, "age"))
  expect_equal(docvars(corpus_split, "thematics"), docvars(corpus_split_id, "thematics"))
  expect_equal(docvars(corpus_split, "thematics"), c("t1", "t2", "t1", "t2", "t3", "t4"))
})

text <- "****
-*t1
Donec placerat quis quam ac interdum.
-*t2
Vivamus dictum, nisi et suscipit fermentum, tortor nulla tempor quam, a ullamcorper elit erat nec arcu.
****
-*t1
Vestibulum vitae ante eget justo maximus luctus.
-*t2
Nam nunc neque, fringilla vel pharetra sit amet, posuere a nibh."

con <- textConnection(text)
corpus_without_meta <- import_corpus_iramuteq(con, id_var = "id", thematics = "remove")
close(con)
con <- textConnection(text)
corpus_without_meta_split <- import_corpus_iramuteq(con, id_var = "id", thematics = "split")
close(con)


test_that("corpus without meta is ok", {
  expect_equal(ndoc(corpus_without_meta), 2)
  expect_equal(as.character(corpus_without_meta)[[2]], c("Vestibulum vitae ante eget justo maximus luctus.\nNam nunc neque, fringilla vel pharetra sit amet, posuere a nibh."))
  expect_equal(dim(docvars(corpus_without_meta)), c(2, 0))
})

test_that("splitted corpus without meta is ok", {
  expect_equal(dim(docvars(corpus_without_meta_split)), c(4, 1))
  expect_equal(docvars(corpus_without_meta_split, "thematics"), c("t1","t2","t1","t2"))
  expect_equal(ndoc(corpus_without_meta_split), 4)
  expect_equal(as.character(corpus_without_meta_split)[[3]], "Vestibulum vitae ante eget justo maximus luctus.")
})

text <- "****
Donec placerat quis quam ac interdum.
**** *foo_1
Vestibulum vitae ante eget justo maximus luctus.
****
Donec placerat quis quam ac interdum.
**** *bar_bar
Vestibulum vitae ante eget justo maximus luctus."

con <- textConnection(text)
corpus_with_partial_meta <- import_corpus_iramuteq(con, id_var = "id", thematics = "remove")
close(con)

test_that("corpus with partial meta is ok", {
  expect_equal(as.character(corpus_with_partial_meta)[[2]], "Vestibulum vitae ante eget justo maximus luctus.")
  expect_equal(ndoc(corpus_with_partial_meta), 4)
  expect_equal(dim(docvars(corpus_with_partial_meta)), c(4, 2))
  expect_equal(docvars(corpus_with_partial_meta, "foo"), c(NA, "1", NA, NA))
  expect_equal(docvars(corpus_with_partial_meta, "bar"), c(NA, NA, NA, "bar"))
})

## CSS and modules for Shiny gadgets

## CSS for shiny gadgets

rainette_explor_css <- function() {
  "
#main {
  padding: 0 1em;
}
#docs {
  height: 100%;
  max-height: 100%;
  overflow-y: hidden;
  padding-left: 2em;
}
#docs_sample {
  height: 100%;
  max-height: 100%;
  overflow-y: scroll;
  padding-right: 2em;
}
#docs_sample_intro {
  color: #369;
  background-color: #F0F0F0;
  border-radius: 5px;
  padding: .6em 1em;
  margin-bottom: 20px;
}
#docs_sample hr {
  margin-top: 12px;
  margin-bottom: 12px;
}
#side {
  background-color: #EEEEEE;
  padding: 2em 3em;
  font-size: 12px;
}
.docname {
  font-size: 80%;
  color: #69B;
  margin: 0 0 .3em 0;
}
.doc {
  font-size: 100%;
  max-width: 50em;
  border-left: 3px solid #9BE;
  margin: 0;
  padding: .3em 1em .2em 1em;
}
.doc .highlight {
  background-color: #FF0;
}

/* Syntax highlighting */
span.hl.str { color: #d14;}
span.hl.kwa { color: #099;}
span.hl.num { color: #099;}
span.hl.kwd { color: #333; font-weight: bold;}
span.hl.com { color: #888; font-style: italic;}
"
}


## Cluster documents sample UI

docs_sample_ui <- function(id, res) {
  show_merged_segments <- !inherits(res, "rainette2") &&
    !is.null(res$call$min_segment_size) &&
    res$call$min_segment_size > 1

  ns <- shiny::NS(id)
  fillRow(
    flex = c(1, 3),
    fillCol(
      flex = c(10, 1),
      id = "side",
      div(
        uiOutput(ns("group_ui")),
        numericInput(
          ns("ndoc"),
          "Documents displayed",
          value = 100,
          min = 1
        ),
        checkboxInput(ns("random_sample"), "Random sample", value = FALSE),
        numericInput(
          ns("nchar"),
          "Maximum text length",
          value = 1000,
          min = 10
        ),
        textInput(
          ns("filter_term"),
          "Filter by term",
          value = ""
        ),
        if (show_merged_segments) {
          shiny::checkboxInput(
            ns("show_merged"),
            "Show merged segments",
            value = FALSE
          )
        }
      )
    ),
    fillCol(
      flex = c(10, 1),
      id = "docs",
      fillCol(
        flex = c(NA, 1),
        div(id = "docs_sample_intro", htmlOutput(ns("docs_sample_intro"))),
        div(id = "docs_sample", htmlOutput(ns("docs_sample")))
      )
    )
  )
}


## Cluster documents sample Server

docs_sample_server <- function(id, res, corpus_src, current_k) {
  moduleServer(
    id,
    function(input, output, session) {
      ## Cluster selection slider
      output$group_ui <- renderUI({
        ns <- session$ns
        selectInput(
          ns("cluster"),
          label = "Cluster",
          selected = 1,
          choices = seq_len(current_k())
        )
      })

      ## Current clusters
      groups <- reactive({
        rainette::cutree(res, k = current_k())
      })

      ## Corpus from wanted cluster
      corpus_cluster <- reactive({
        if (is.null(corpus_src)) {
          return(NULL)
        }
        ## Select only from wanted cluster
        if (!is.null(input$show_merged) && input$show_merged) {
          corpus_src$group <- groups()
          corpus_src$doc_name <- quanteda::docnames(corpus_src)
          corpus_src$uc_id <- res$corresp_uce_uc$uc
          docvars(corpus_src) <- docvars(corpus_src) %>%
            group_by(.data$uc_id) %>%
            mutate(doc_name = paste(.data$doc_name, collapse = " | ")) %>%
            ungroup()
          result <- quanteda::corpus_group(
            corpus_src,
            groups = corpus_src$doc_name
          )
          sel <- quanteda::docvars(result, "group") == input$cluster &
            !is.na(quanteda::docvars(result, "group"))
          result <- result[sel]
        } else {
          sel <- groups() == input$cluster & !is.na(groups())
          result <- corpus_src[sel]
        }
        result
      })

      ## Regex for filter term input
      filter_regex <- reactive({
        stringr::regex(
          shiny::req(input$filter_term),
          ignore_case = TRUE,
          multiline = TRUE
        )
      })

      ## Corpus from cluster filtered by term
      corpus_filtered <- reactive({
        if (is.null(corpus_src)) {
          return(NULL)
        }
        result <- corpus_cluster()
        ## Filter by terms
        filter_term <- stringr::str_trim(input$filter_term)
        if (!is.null(filter_term) && filter_term != "") {
          keep <- stringr::str_detect(
            as.character(corpus_cluster()),
            filter_regex()
          )
          result <- result[keep]
        }
        result
      })

      highlighter <- reactive({
        filter_term <- stringr::str_trim(input$filter_term)
        if (!is.null(filter_term) && filter_term != "") {
          fun <- function(txt) {
            stringr::str_replace_all(
              txt,
              filter_regex(),
              "<span class='highlight'>\\0</span>"
            )
          }
        } else {
          fun <- I
        }
        fun
      })

      ## Maximum number of documents to display
      n_doc <- reactive({
        n_doc <- input$ndoc
        if (is.null(n_doc) || is.na(n_doc) || n_doc < 1) {
          n_doc <- 1
        }
        n_doc
      })

      ## Maximum number of characters per document
      n_char <- reactive({
        n_char <- input$nchar
        if (is.null(n_char) || is.na(n_char) || n_char < 1) {
          n_char <- 1
        }
        n_char
      })

      ## Sample cluster documents introducation phrase
      output$docs_sample_intro <- renderUI({
        if (is.null(corpus_src)) {
          return(
            HTML(
              "<p>Can't display documents : <tt>corpus_src</tt> is null.</p><p>Please rerun <tt>rainette_explor</tt> with your quanteda corpus object as third parameter : something like <tt>rainette_explor(res, dtm, corpus)</tt>.</p>"
            )
          )
        }

        nb_docs_cluster <- quanteda::ndoc(corpus_filtered())
        out <- paste0(
          "Displayed : <strong>",
          min(n_doc(), nb_docs_cluster),
          "</strong>"
        )
        if (
          quanteda::ndoc(corpus_cluster()) != quanteda::ndoc(corpus_filtered())
        ) {
          out <- paste0(
            out,
            " - Filtered documents : <strong>",
            quanteda::ndoc(corpus_filtered()),
            "</strong>"
          )
        }
        out <- paste0(
          out,
          " - Cluster size : <strong>",
          quanteda::ndoc(corpus_cluster()),
          "</strong>."
        )
        shiny::HTML(out)
      })

      ## Sample cluster documents
      output$docs_sample <- renderUI({
        if (is.null(corpus_src)) {
          return(NULL)
        }

        ## Sample docs
        size <- min(quanteda::ndoc(corpus_filtered()), n_doc())
        if (input$random_sample) {
          corp <- quanteda::corpus_sample(corpus_filtered(), size = size)
        } else {
          corp <- corpus_filtered()[seq_len(size)]
        }

        ## Truncate texts
        txt <- as.character(corp)
        txt <- ifelse(
          nchar(txt) <= n_char(),
          txt,
          paste(stringr::str_sub(txt, 1, n_char()), "(...)")
        )

        ## Highlight texts
        txt <- highlighter()(txt)

        ## Generate output
        out <- paste(
          "<div class='doc'>",
          "<div class='docname'>",
          quanteda::docnames(corp),
          "</div>",
          stringr::str_replace_all(txt, "\n", "<br>"),
          "</div>",
          "<hr>",
          collapse = "\n"
        )
        shiny::HTML(out)
      })
    }
  )
}

#' Merges segments according to minimum segment size
#'
#' `rainette_uc_index` docvar
#'
#' @param dtm dtm of segments
#' @param min_segment_size minimum number of forms by segment
#' @param doc_id character name of a dtm docvar which identifies source documents.
#'
#' @details
#' If `min_segment_size == 0`, no segments are merged together.
#' If `min_segment_size > 0` then `doc_id` must be provided
#' unless the corpus comes from `split_segments`, in this case
#' `segment_source` is used by default.
#'
#' @return the original dtm with a new `rainette_uc_id` docvar.
#' @export

merge_segments <- function(dtm, min_segment_size = 10, doc_id = NULL) {
  ## Add id to documents
  quanteda::docvars(dtm, field = "rainette_uce_id") <- seq_len(nrow(dtm))

  if (min_segment_size <= 1) {
    ## Do nothing
    quanteda::docvars(dtm, field = "rainette_uc_id") <- quanteda::docvars(
      dtm,
      "rainette_uce_id"
    )
    return(dtm)
  }

  ## Check for min_segment_size and doc_id values
  if (is.null(doc_id)) {
    if ("segment_source" %in% names(docvars(dtm))) {
      doc_id <- "segment_source"
    } else {
      stop("If min_segment_size > 0, you must provide a doc_id value.")
    }
  }

  ## Size of each uce
  terms_by_uce <- rowSums(dtm)
  doc_ids <- quanteda::docvars(dtm, doc_id)

  ## If all uce are already above the minimum size
  if (all(terms_by_uce >= min_segment_size)) {
    quanteda::docvars(dtm, "rainette_uc_id") <- quanteda::docvars(
      dtm,
      "rainette_uce_id"
    )
    return(dtm)
  }

  ## else
  index <- 1
  uc_id <- quanteda::docvars(dtm, "rainette_uce_id")
  while (index <= length(terms_by_uce)) {
    current_size <- terms_by_uce[index]
    grouping_index <- index
    ## While current uc size is smaller than min, regroup with following uce
    while (current_size < min_segment_size) {
      if (
        (grouping_index + 1) <= length(terms_by_uce) &&
          doc_ids[grouping_index] == doc_ids[grouping_index + 1]
      ) {
        grouping_index <- grouping_index + 1
        current_size <- current_size + terms_by_uce[grouping_index]
        uc_id[grouping_index] <- index
      } else {
        ## If new index is out of bounds or in another document
        ## replace current uc index with the previous one, if any
        current_doc_id <- doc_ids[grouping_index]
        current_uc_id <- uc_id[grouping_index]
        other_uc_ids <- uc_id[doc_ids == current_doc_id & uc_id < current_uc_id]
        if (length(other_uc_ids) > 0) {
          previous_uc_id <- max(other_uc_ids)
          uc_id[uc_id == current_uc_id] <- previous_uc_id
        }
        break
      }
    }
    index <- grouping_index + 1
  }
  ## Add computed uc ids to docvars
  quanteda::docvars(dtm, "rainette_uc_id") <- uc_id

  ## Test if any uc is below min_segment_size
  dtm_uc_size <- quanteda::dfm_group(
    dtm,
    quanteda::docvars(dtm, "rainette_uc_id")
  )
  if (any(rowSums(dtm_uc_size) < min_segment_size)) {
    warning("some segments will have a size < min_segment_size")
  }

  return(dtm)
}


#' Returns the number of segment of each cluster for each source document
#'
#' @param obj a corpus, tokens or dtm object
#' @param clust_var name of the docvar with the clusters
#' @param doc_id docvar identifying the source document
#' @param prop if TRUE, returns the percentage of each cluster by document
#'
#' @details
#' This function is only useful for previously segmented corpus. If `doc_id`
#' is NULL and there is a `sement_source` docvar, it will be used instead.
#'
#' @seealso [docs_by_cluster_table()]
#'
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 2)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' corpus$cluster <- cutree(res, k = 3)
#' clusters_by_doc_table(corpus, clust_var = "cluster", prop = TRUE)
#' }
#' @export

clusters_by_doc_table <- function(
  obj,
  clust_var = NULL,
  doc_id = NULL,
  prop = FALSE
) {
  if (
    !inherits(obj, "corpus") &&
      !inherits(obj, "dfm") &&
      !inherits(obj, "tokens")
  ) {
    stop("obj must be a corpus, a tokens or a dfm object.")
  }

  if (is.null(doc_id) && "segment_source" %in% names(docvars(obj))) {
    doc_id <- "segment_source"
  }

  ids <- quanteda::docvars(obj, doc_id)
  res <- dplyr::tibble(
    doc_id = factor(ids, levels = unique(ids)),
    cluster = docvars(obj, clust_var)
  )

  names_prefix <- ""
  if (is.numeric(res$cluster)) {
    res$cluster <- as.character(res$cluster)
    names_prefix <- "clust_"
  }

  ## Convert NA to "NA" to keep them if there is not already "NA" values
  if (any(is.na(res$cluster))) {
    if (!("NA" %in% res$cluster)) {
      res$cluster[is.na(res$cluster)] <- "NA"
    } else {
      res$cluster[is.na(res$cluster)] <- "NA_missing_"
    }
  }

  ## Count clusters
  res <- res %>%
    dplyr::count(.data$doc_id, .data$cluster)

  ## Compute percenteages
  if (prop) {
    res <- res %>%
      dplyr::group_by(.data$doc_id) %>%
      dplyr::mutate(n = .data$n / sum(.data$n) * 100) %>%
      dplyr::ungroup()
  }

  ## Pivoting
  res <- res %>%
    tidyr::pivot_wider(
      id_cols = "doc_id",
      names_from = "cluster",
      values_from = "n",
      names_prefix = names_prefix,
      values_fill = 0
    ) %>%
    dplyr::mutate(doc_id = as.character(.data$doc_id))

  cols <- sort(colnames(res))
  cols <- cols[cols != "doc_id"]
  dplyr::relocate(res, doc_id, all_of(cols))
}


#' Returns, for each cluster, the number of source documents with at least n
#' segments of this cluster
#'
#' @param obj a corpus, tokens or dtm object
#' @param clust_var name of the docvar with the clusters
#' @param doc_id docvar identifying the source document
#' @param threshold the minimal number of segments of a given cluster that a document
#'   must include to be counted
#'
#' @details
#' This function is only useful for previously segmented corpus. If `doc_id` is NULL
#' and there is a `sement_source` docvar, it will be used instead.
#'
#' @seealso [clusters_by_doc_table()]
#'
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 2)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' corpus$cluster <- cutree(res, k = 3)
#' docs_by_cluster_table(corpus, clust_var = "cluster")
#' }
#' @export

docs_by_cluster_table <- function(
  obj,
  clust_var = NULL,
  doc_id = NULL,
  threshold = 1
) {
  count <- clusters_by_doc_table(
    obj,
    clust_var = clust_var,
    doc_id = doc_id,
    prop = FALSE
  )

  count %>%
    dplyr::select(-"doc_id") %>%
    dplyr::mutate(dplyr::across(everything(), .fns = function(v) {
      v >= threshold
    })) %>%
    dplyr::summarise(dplyr::across(everything(), .fns = sum)) %>%
    tidyr::pivot_longer(
      cols = dplyr::everything(),
      names_to = "cluster",
      values_to = "n"
    ) %>%
    dplyr::mutate(`%` = .data$n / nrow(.env$count) * 100)
}


#' @importFrom rlang sym

stat_col <- function(measure) {
  switch(
    measure,
    "chi2" = "chi2",
    "lr" = "G2",
    "frequency" = "frequency",
    "docprop" = "docprop"
  )
}

library(quanteda)
context("utils functions")

m <- matrix(
  c(1, 1, 1, 0, 0,
    0, 0, 1, 1, 0,
    1, 0, 0, 0, 0,
    1, 1, 1, 1, 0,
    1, 1, 1, 1, 1,
    1, 0, 0, 0, 0,
    1, 1, 1, 0, 0
    ),
  ncol = 5,
  byrow = TRUE,
)
mini_dfm <- as.dfm(m)
docvars(mini_dfm, "rainette_uce_id") <- seq_len(nrow(mini_dfm))
docvars(mini_dfm, "doc_id") <- "doc1"


# merge_segments

test_that("computed uc are ok when min_segment_size == 0", {
  dfm_uc <- rainette::merge_segments(mini_dfm, min_segment_size = 0)
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 3, 4, 5, 6, 7))
})

test_that("computed uc are ok when min_segment_size == 1", {
  dfm_uc <- rainette::merge_segments(mini_dfm, min_segment_size = 1)
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 3, 4, 5, 6, 7))
})

test_that("error when min_segment_size > 1 and no doc_id", {
  expect_error(merge_segments(mini_dfm, min_segment_size = 3))
})

test_that("computed uc are ok when min_segment_size > 0", {
  dfm_uc <- rainette::merge_segments(mini_dfm, min_segment_size = 3, doc_id = "doc_id")
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 2, 4, 5, 6, 6))
})

test_that("warning if uc can't be computed", {
  expect_warning(merge_segments(mini_dfm, min_segment_size = 20, doc_id = "doc_id"))
})

mini_dfm_docs <- mini_dfm
test_that("computed uc are ok with respect to doc_id", {
  docvars(mini_dfm_docs, "doc_id") <- c("doc1", "doc1", "doc1", "doc2", "doc2", "doc2", "doc3")
  dfm_uc <- rainette::merge_segments(mini_dfm_docs, min_segment_size = 3, doc_id = "doc_id")
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 2, 4, 5, 5, 7))

  docvars(mini_dfm_docs, "doc_id") <- c("doc1", "doc1", "doc1", "doc2", "doc3", "doc3", "doc3")
  dfm_uc <- rainette::merge_segments(mini_dfm_docs, min_segment_size = 4, doc_id = "doc_id")
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 1, 1, 4, 5, 6, 6))

  docvars(mini_dfm_docs, "doc_id") <- c("doc1", "doc1", "doc1", "doc1", "doc3", "doc3", "doc3")
  dfm_uc <- rainette::merge_segments(mini_dfm_docs, min_segment_size = 6, doc_id = "doc_id")
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 1, 1, 1, 5, 5, 5))

  docvars(mini_dfm_docs, "doc_id") <- c("doc1", "doc2", "doc3", "doc4", "doc4", "doc5", "doc5")
  expect_warning(dfm_uc <- rainette::merge_segments(mini_dfm_docs, min_segment_size = 3, doc_id = "doc_id"))
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 3, 4, 5, 6, 6))

  corpus <- head(data_corpus_inaugural, n = 10)
  corpus <- split_segments(corpus)
  dtm <- dfm(tokens(corpus, remove_punct = TRUE), tolower = TRUE)
  dtm <- dfm_remove(dtm, stopwords("en"))
  dtm <- dfm_trim(dtm, min_termfreq = 3)
  expect_warning(dfm_uc <- rainette::merge_segments(dtm, min_segment_size = 500))
  tmp <- docvars(dfm_uc) %>%
    group_by(rainette_uc_id) %>%
    summarise(nd = n_distinct(segment_source)) %>%
    filter(nd > 1)
  expect_equal(nrow(tmp), 0)
})

test_that("warning if uc can't be computed with respect to doc_id", {
  docvars(mini_dfm_docs, "doc_id") <- c("doc1", "doc1", "doc1", "doc2", "doc2", "doc2", "doc3")
  expect_warning(merge_segments(mini_dfm_docs, min_segment_size = 4, doc_id = "doc_id"))
})

mini_dfm_segsource <- mini_dfm
docvars(mini_dfm_segsource, "segment_source") <- "doc1"
test_that("segment_source used by default when min_segment_size > 0", {
  dfm_uc <- rainette::merge_segments(mini_dfm_segsource, min_segment_size = 3)
  expect_equal(docvars(dfm_uc, "rainette_uc_id"), c(1, 2, 2, 4, 5, 6, 6))
})


# clusters_by_doc_table
mini_dfm_count <- mini_dfm
docvars(mini_dfm_count, "segment_source") <- c("doc1", "doc1", "doc2", "doc2", "doc2", "doc2", "doc3")
docvars(mini_dfm_count, "doc_id") <- docvars(mini_dfm_count, "segment_source")
test_that("clusters_by_doc_table results", {
  docvars(mini_dfm_count, "cluster") <- c(1, 2, 2, 1, 2, NA, 1)
  expect_equal(
    clusters_by_doc_table(mini_dfm_count, clust_var = "cluster"),
    structure(list(doc_id = c("doc1", "doc2", "doc3"), clust_1 = c(1L,
      1L, 1L), clust_2 = c(1L, 2L, 0L), clust_NA = c(0L, 1L, 0L)), row.names = c(NA,
      -3L), class = c("tbl_df", "tbl", "data.frame"))
  )
  expect_equal(
    clusters_by_doc_table(mini_dfm_count, clust_var = "cluster", prop = TRUE),
    structure(list(doc_id = c("doc1", "doc2", "doc3"), clust_1 = c(50,
      25, 100), clust_2 = c(50, 50, 0), clust_NA = c(0, 25, 0)), row.names = c(NA,
      -3L), class = c("tbl_df", "tbl", "data.frame"))
  )
  expect_equal(
    clusters_by_doc_table(mini_dfm_count, clust_var = "cluster", doc_id = "doc_id", prop = TRUE),
    structure(list(doc_id = c("doc1", "doc2", "doc3"), clust_1 = c(50,
      25, 100), clust_2 = c(50, 50, 0), clust_NA = c(0, 25, 0)), row.names = c(NA,
      -3L), class = c("tbl_df", "tbl", "data.frame"))
  )
  docvars(mini_dfm_count, "cluster") <- c("C1", "C2", "C2", "C1", "NA", NA, NA)
  expect_equal(
    clusters_by_doc_table(mini_dfm_count, clust_var = "cluster"),
    structure(list(doc_id = c("doc1", "doc2", "doc3"), C1 = c(1L,
      1L, 0L), C2 = c(1L, 1L, 0L), `NA` = c(0L, 1L, 0L), NA_missing_ = c(0L,
      1L, 1L)), row.names = c(NA, -3L), class = c("tbl_df", "tbl",
      "data.frame"))
  )
})

# docs_by_cluster_table
test_that("docs_by_cluster_table results", {
  docvars(mini_dfm_count, "cluster") <- c(1, 2, 2, 1, 2, NA, 1)
  expect_equal(
    docs_by_cluster_table(mini_dfm_count, clust_var = "cluster"),
    structure(list(cluster = c("clust_1", "clust_2", "clust_NA"),
      n = 3:1, `%` = c(100, 66.6666666666667, 33.3333333333333)), row.names = c(NA,
      -3L), class = c("tbl_df", "tbl", "data.frame"))
  )
  expect_equal(
    docs_by_cluster_table(mini_dfm_count, clust_var = "cluster", threshold = 2),
    structure(list(cluster = c("clust_1", "clust_2", "clust_NA"),
      n = c(0, 1, 0), `%` = c(0, 33.3333333333333, 0)), row.names = c(NA,
      -3L), class = c("tbl_df", "tbl", "data.frame"))
  )
  docvars(mini_dfm_count, "cluster") <- c("C1", "C2", "C2", "C1", "NA", NA, NA)
  expect_equal(
    docs_by_cluster_table(mini_dfm_count, clust_var = "cluster", doc_id = "doc_id"),
    structure(list(cluster = c("C1", "C2", "NA", "NA_missing_"),
      n = c(2L, 2L, 1L, 2L), `%` = c(66.6666666666667, 66.6666666666667,
      33.3333333333333, 66.6666666666667)), row.names = c(NA, -4L
      ), class = c("tbl_df", "tbl", "data.frame"))
  )
})

# stat_col

test_that("stat_col", {
  expect_equal(stat_col("chi2"), "chi2")
  expect_equal(stat_col("lr"), "G2")
})

##' Split a character string or corpus into segments
##'
##' Split a character string or corpus into segments, taking into account punctuation where possible
##'
##' @param obj character string, quanteda or tm corpus object
##' @param segment_size segment size (in words)
##' @param segment_size_window window around segment size to look for best splitting point
##'
##' @return
##' If obj is a tm or quanteda corpus object, the result is a quanteda corpus.
##' @export
##'
##' @examples
##' \donttest{
##' require(quanteda)
##' split_segments(data_corpus_inaugural)
##' }

split_segments <- function(obj, segment_size = 40, segment_size_window = NULL) {
  UseMethod("split_segments")
}


##' @rdname split_segments
##' @aliases split_segments.character
##'
##' @export
##' @import dplyr
##' @import quanteda
##' @importFrom purrr map_chr

split_segments.character <- function(
  obj,
  segment_size = 40,
  segment_size_window = NULL
) {
  ## Tokenize into words
  words <- as.character(quanteda::tokens(obj, what = "fastestword"))

  split_segments_words(words, segment_size, segment_size_window)
}


## Split a single tokenized document
split_segments_words <- function(
  words,
  segment_size = 40,
  segment_size_window = NULL
) {
  if (inherits(words, "tokens") && length(words) == 1) {
    words <- as.character(words)
  }

  ## Default segment_size_window
  if (is.null(segment_size_window)) {
    segment_size_window <- 0.4 * segment_size
  }

  ## If string is shorter than segment_size, returns it
  if (length(words) <= segment_size) {
    return(paste0(words, collapse = " "))
  }

  ## Compute "weight" for each word
  last_char <- stringr::str_sub(words, -1) # nolint
  weights <- dplyr::case_when(
    last_char %in% c(".", "?", "!", "\u2026") ~ 6,
    last_char == ":" ~ 5,
    last_char == ";" ~ 4,
    last_char == "," ~ 1,
    TRUE ~ 0.01
  )

  slice_indices <- 1:(segment_size + segment_size_window)
  last_index <- 1
  split_indices <- 1
  stop_index <- length(words) - (segment_size + segment_size_window) + 1
  while (last_index < stop_index) {
    indices <- last_index + slice_indices - 1
    tmpw <- weights[indices]
    split_index <- which.max(tmpw / (abs(slice_indices - segment_size) + 1))
    split_indices <- append(split_indices, split_index)
    last_index <- last_index + split_index
  }
  split_indices <- append(split_indices, length(words) - sum(split_indices) + 1)
  split_indices <- cumsum(split_indices)

  segments <- purrr::map_chr(
    seq_len(length(split_indices) - 1),
    ~ {
      w <- words[split_indices[.x]:(split_indices[.x + 1] - 1)]
      paste0(w, collapse = " ")
    }
  )

  segments
}


##' @rdname split_segments
##' @aliases split_segments.Corpus
##' @export

split_segments.Corpus <- function(
  obj,
  segment_size = 40,
  segment_size_window = NULL
) {
  corpus <- obj

  if (!inherits(corpus, "Corpus")) {
    stop("corpus must be of class Corpus")
  }

  corpus <- quanteda::corpus(corpus)
  split_segments(corpus, segment_size, segment_size_window)
}


##' @rdname split_segments
##' @aliases split_segments.corpus
##' @export
##' @importFrom purrr map_int

split_segments.corpus <- function(
  obj,
  segment_size = 40,
  segment_size_window = NULL
) {
  if (!inherits(obj, "corpus")) {
    stop("obj must be of class corpus")
  }

  tokens <- quanteda::tokens(obj, what = "fastestword")

  split_segments(tokens, segment_size, segment_size_window)
}


##' @rdname split_segments
##' @aliases split_segments.corpus
##' @export
##' @importFrom purrr map_int

split_segments.tokens <- function(
  obj,
  segment_size = 40,
  segment_size_window = NULL
) {
  if (!inherits(obj, "tokens")) {
    stop("obj must be of class tokens")
  }

  quanteda::docvars(obj, "segment_source") <- quanteda::docnames(obj)

  message("  Splitting...")

  progressr::with_progress({
    p <- progressr::progressor(steps = quanteda::ndoc(obj))

    texts <- lapply(
      obj,
      function(words) {
        p()
        split_segments_words(words, segment_size, segment_size_window)
      }
    )
  })

  new_corpus <- docvars(obj) %>%
    dplyr::mutate(text = texts) %>%
    tidyr::unnest("text") %>%
    dplyr::group_by(.data$segment_source) %>%
    dplyr::mutate(
      segment_id = paste0(.data$segment_source, "_", seq_len(dplyr::n()))
    ) %>%
    quanteda::corpus(
      docid_field = "segment_id",
      text_field = "text"
    )

  message("  Done.")

  new_corpus
}

library(testthat)
library(quanteda)
context("stats functions")

mini_corpus <- head(data_corpus_inaugural, n = 10)
mini_corpus <- split_segments(mini_corpus)
dtm <- dfm(tokens(mini_corpus, remove_punct = TRUE), tolower = TRUE)
dtm <- dfm_remove(dtm, stopwords("en"))
dtm <- dfm_trim(dtm, min_termfreq = 3)

res <- rainette(dtm, k = 5, min_segment_size = 5, min_split_members = 3)

test_that("rainette_stats", {

  ## chi2
  groups <- cutree_rainette(res, k = 3)
  stats <- rainette_stats(groups, dtm, n_terms = 10)
  expect_equal(length(stats), max(groups))
  n_stats_pos <- sum(stats[[2]]$sign == "positive")
  expect_equivalent(
    stats[[2]][stats[[2]]$sign == "positive", -ncol(stats[[2]])],
    head(quanteda.textstats::textstat_keyness(dtm, groups == 2), n_stats_pos)
  )
  n_stats_neg <- sum(stats[[3]]$sign == "negative")
  expect_equivalent(
    stats[[3]][stats[[3]]$sign == "negative",  -ncol(stats[[3]])] %>%
      dplyr::arrange(desc(chi2)),
    tail(quanteda.textstats::textstat_keyness(dtm, groups == 3), n_stats_neg)
  )

  ## lr
  groups <- cutree_rainette(res, k = 4)
  stats <- rainette_stats(groups, dtm, measure = "lr", n_terms = 8)
  expect_equal(length(stats), max(groups))
  n_stats_pos <- sum(stats[[1]]$sign == "positive")
  expect_equivalent(
    stats[[1]][stats[[1]]$sign == "positive", -ncol(stats[[1]])],
    head(quanteda.textstats::textstat_keyness(dtm, measure = "lr", groups == 1), n_stats_pos)
  )
  n_stats_neg <- sum(stats[[4]]$sign == "negative")
  expect_equivalent(
    stats[[4]][stats[[4]]$sign == "negative", -ncol(stats[[4]])] %>%
      dplyr::arrange(desc(G2)),
    tail(quanteda.textstats::textstat_keyness(dtm, measure = "lr", groups == 4), n_stats_neg)
  )

  ## max_p
  stats <- rainette_stats(groups, dtm, n_terms = 10, max_p = 1e-09)
  for (i in seq_along(stats)) {
    expect(all(stats[[i]]$p <= 1e-09), "p not less than max_p")
  }

  ## frequency and docprop
  groups <- cutree_rainette(res, k = 3)
  stats <- rainette_stats(groups, dtm, n_terms = 10, measure = "frequency")
  expect_equal(length(stats), max(groups))
  dtm2 <- dfm_subset(dtm, groups == 2)
  expect_equivalent(
    stats[[2]] %>% select(-sign),
    quanteda.textstats::textstat_frequency(dtm2) %>%
      as_tibble() %>%
      mutate(docprop = .data$docfreq / ndoc(dtm2)) %>%
      arrange(desc(frequency)) %>%
      slice(1:10)
  )
  stats <- rainette_stats(groups, dtm, n_terms = 10, measure = "docprop")
  expect_equal(length(stats), max(groups))
  dtm3 <- dfm_subset(dtm, groups == 3)
  expect_equivalent(
    stats[[3]] %>% select(-sign),
    quanteda.textstats::textstat_frequency(dtm3) %>%
      as_tibble() %>%
      mutate(docprop = .data$docfreq / ndoc(dtm3)) %>%
      arrange(desc(docprop)) %>%
      head(10)
  )
})

library(quanteda)

context("split_segments")


data(data_corpus_inaugural)
corpus <- head(data_corpus_inaugural)
text <- as.character(corpus)[[1]]

split_char <- split_segments(text)

split_quanteda <- split_segments(corpus)

test_that("split_segments.character is ok", {
  expect_equal(length(split_char), 37)
  expect_equal(split_char[1], "Fellow-Citizens of the Senate and of the House of Representatives:\n\nAmong the vicissitudes incident to life no event could have filled me with greater anxieties than that of which the notification was transmitted by your order, and received on the 14th day of the present month.")
  expect_equal(split_segments("One more test."), "One more test.")
  expect_equal(split_segments("J'apprends des techniques d'écriture. Cela me permet d'écrire autre choses que des lettres ou un récit de ma vie/journal_intime. Cela donne des idées pour écrire des fictions, sortir de son quotidien. J'aime bien l'effort intellectuel que cela demande. C'est un rendez-vous apaisant."), "J'apprends des techniques d'écriture. Cela me permet d'écrire autre choses que des lettres ou un récit de ma vie/journal_intime. Cela donne des idées pour écrire des fictions, sortir de son quotidien. J'aime bien l'effort intellectuel que cela demande. C'est un rendez-vous apaisant.")

  expect_equal(
    split_segments("Cool ! Cool cool cool."),
    "Cool ! Cool cool cool."
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 1),
    c("Cool", "!", "Cool", "cool", "cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 1, 1),
    c("Cool !", "Cool", "cool cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 1, 2),
    c("Cool !", "Cool cool cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 2),
    c("Cool !", "Cool cool", "cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 2, 2),
    c("Cool !", "Cool cool cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 3),
    c("Cool !", "Cool cool cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 3, 2),
    c("Cool ! Cool cool cool.")
  )
  expect_equal(
    split_segments("Cool ! Cool cool cool.", 3, 3),
    "Cool ! Cool cool cool."
  )
})

test_that("split_segments.Corpus is ok", {
  skip_if_not_installed("tm")
  if (requireNamespace("tm", quietly = TRUE)) {
    data(acq, package = "tm")
    split_tm <- split_segments(acq)
    expect_equal(ndoc(split_tm), 188)
    skip_if(utils::packageVersion("quanteda") < "2.0.0")
    expect_equal(docvars(split_tm, "segment_source")[1], "reut-00001.xml")
    expect_equal(docnames(split_tm)[1], "reut-00001.xml_1")
  }
})

test_that("split_segments.corpus is ok", {
  expect_equal(ndoc(split_quanteda), 229)
  expect_equal(docvars(split_quanteda, "segment_source")[1], "1789-Washington")
  expect_equal(docnames(split_quanteda)[1], "1789-Washington_1")
})

test_that("split_segments input checking is ok", {
  expect_error(rainette:::split_segments.character(1:10))
  expect_error(rainette:::split_segments.corpus(1:10))
  expect_error(rainette:::split_segments.Corpus(1:10))
})

#' Corpus clustering based on the Reinert method - Simple clustering
#'
#' @param dtm quanteda dfm object of documents to cluster, usually the
#'   result of [split_segments()]
#' @param k maximum number of clusters to compute
#' @param min_segment_size minimum number of forms by document
#' @param doc_id character name of a dtm docvar which identifies source documents.
#' @param min_split_members don't try to split groups with fewer members
#' @param cc_test contingency coefficient value for feature selection
#' @param tsj minimum frequency value for feature selection
#' @param min_members deprecated, use `min_split_members` instead
#' @param min_uc_size deprecated, use `min_segment_size` instead
#'
#' @details
#' See the references for original articles on the method. Computations and results may differ
#' quite a bit, see the package vignettes for more details.
#'
#' The dtm object is automatically converted to boolean.
#'
#' If `min_segment_size > 0` then `doc_id` must be provided unless the corpus comes from `split_segments`,
#' in this case `segment_source` is used by default.
#'
#' @return
#' The result is a list of both class `hclust` and `rainette`. Besides the elements
#' of an `hclust` object, two more results are available :
#'
#' - `uce_groups` give the group of each document for each k
#' - `group` give the group of each document for the maximum value of k available
#'
#' @seealso [split_segments()], [rainette2()], [cutree_rainette()], [rainette_plot()], [rainette_explor()]
#'
#' @references
#'
#' - Reinert M, Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte, Cahiers de l'analyse des données, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
#' - Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Méthodologie Sociologique, Volume 26, Numéro 1, 1990. \doi{10.1177/075910639002600103}
#'
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 3)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' }
#'
#' @export

rainette <- function(
  dtm,
  k = 10,
  min_segment_size = 0,
  doc_id = NULL,
  min_split_members = 5,
  cc_test = 0.3,
  tsj = 3,
  min_members,
  min_uc_size
) {
  ## Check for deprecated min_members argument
  if (!missing(min_members)) {
    warning("min_members is deprecated. Use min_split_members instead.")
    if (missing(min_split_members)) {
      min_split_members <- min_members
    }
  }
  ## Check for deprecated min_uc_size argument
  if (!missing(min_uc_size)) {
    warning("min_uc_size is deprecated. Use min_segment_size instead.")
    if (missing(min_segment_size)) {
      min_segment_size <- min_uc_size
    }
  }

  if (min_split_members < 3) {
    warning("min_split_members has been set to 3, its smallest possible value.")
    min_split_members <- 3
  }

  ## Remove empty strings to avoid subcript out of bounds errors
  dtm <- quanteda::dfm_remove(dtm, "")
  dtm <- quanteda::dfm_weight(dtm, scheme = "boolean")

  if (min_segment_size > 1) {
    ## Compute uc from uces based on minimum size
    message("  Merging segments to comply with min_segment_size...")
  }
  dtm <- rainette::merge_segments(
    dtm,
    min_segment_size = min_segment_size,
    doc_id = doc_id
  )

  ## Correspondance table between uce and uc
  corresp_uce_uc <- data.frame(
    uce = docvars(dtm, "rainette_uce_id"),
    uc = docvars(dtm, "rainette_uc_id")
  )
  ## Group dfm by uc
  dtm <- quanteda::dfm_group(dtm, quanteda::docvars(dtm, "rainette_uc_id"))
  dtm <- quanteda::dfm_weight(dtm, scheme = "boolean", force = TRUE)
  if (any(rowSums(dtm) == 0)) {
    warning(
      "some documents don't have any term, they won't be assigned to any cluster."
    )
  }

  message("  Clustering...")
  ## Initialize results list with first dtm
  res <- list(list(tabs = list(dtm)))

  ## Display progress bar
  progressr::with_progress({
    p <- progressr::progressor(along = seq_len(k - 1))

    for (i in 1:(k - 1)) {
      ## Split the biggest group
      biggest_group <- which.max(purrr::map(res[[i]]$tabs, nrow))
      if (nrow(res[[i]]$tabs[[biggest_group]]) < min_split_members) {
        message(
          "! No more group bigger than min_split_members. Stopping after iteration ",
          i,
          "."
        )
        k <- i
        break
      }
      tab <- res[[i]]$tabs[[biggest_group]]
      ## textmodel_ca only works if nrow >= 3 and ncol >= 3
      if (nrow(tab) < 3 || ncol(tab) < 3) {
        message(
          "! Tab to be splitted is not big enough. Stopping after iteration ",
          i,
          "."
        )
        k <- i
        break
      }

      clusters <- cluster_tab(tab, cc_test = cc_test, tsj = tsj)
      p()

      ## Populate results
      res[[i + 1]] <- list()
      res[[i + 1]]$height <- clusters$height
      res[[i + 1]]$splitted <- c(biggest_group, biggest_group + 1)
      res[[i + 1]]$tabs <- append(
        res[[i]]$tabs[-biggest_group],
        clusters$tabs,
        after = biggest_group - 1
      )
      res[[i + 1]]$groups <- append(
        res[[i]]$groups[-biggest_group],
        clusters$groups,
        after = biggest_group - 1
      )
    }
  })

  if (k == 1) {
    message("! No computed clusters. Returning NULL.")
    return(NULL)
  }

  res <- res[-1]

  ## Compute the merge element of resulting hclust result
  groups <- 1:k
  merge <- matrix(nrow = 0, ncol = 2)
  for (i in (k - 1):1) {
    split <- res[[i]]$splitted
    merge <- rbind(merge, -groups[split])
    groups <- groups[-split[2]]
    groups[split[1]] <- -(k - i)
  }

  ## Compute groups by uce at each k
  uce_groups <- list()
  for (i in 1:(k - 1)) {
    group <- rep(NA, nrow(corresp_uce_uc))
    indices <- res[[i]]$groups
    for (group_index in seq_along(indices)) {
      group[corresp_uce_uc$uc %in% indices[[group_index]]] <- group_index
    }
    uce_groups[[i]] <- group
  }

  ## Get the final group element of resulting hclust result
  ## by uce
  group <- uce_groups[[k - 1]]

  message("  Done.")

  ## Compute and return hclust-class result
  hres <- list(
    method = "reinert",
    call = match.call(),
    height = cumsum(rev(purrr::map_dbl(res, ~ .x$height))),
    order = 1:k,
    labels = as.character(1:k),
    merge = merge,
    group = group,
    uce_groups = uce_groups,
    corresp_uce_uc = corresp_uce_uc
  )

  class(hres) <- c("rainette", "hclust")
  hres
}


#' return documents indices ordered by CA first axis coordinates
#'
#' @param m dtm on which to compute the CA and order documents, converted to an integer matrix.
#'
#' @details
#' Internal function, not to be used directly
#'
#' @return ordered list of document indices

order_docs <- function(m) {
  ## Compute first factor of CA on DTM
  ## Code taken from getAnywhere(textmodel_ca.dfm)
  p <- m / sum(m)
  rm <- rowSums(p)
  cm <- colSums(p)
  ep <- tcrossprod(rm, cm)
  s <- (p - ep) / sqrt(ep)
  dec <- RSpectra::svds(s, k = 1, nv = 0)
  coord <- dec$u[, 1] / sqrt(rm)

  ## Order documents by their first factor coordinates
  indices <- (seq_along(coord))[order(coord)]

  return(indices)
}


#' Switch documents between two groups to maximize chi-square value
#'
#' @param m original dtm
#' @param indices documents indices orderes by first CA axis coordinates
#' @param max_index document index where the split is maximum
#' @param max_chisq maximum chi-square value
#'
#' @details
#' Internal function, not to be used directly
#'
#' @return a list of two vectors `indices1` and `indices2`, which contain
#' the documents indices of each group after documents switching, and a `chisq` value,
#' the new corresponding chi-square value after switching

switch_docs <- function(m, indices, max_index, max_chisq) {
  ## Group indices and tabs
  group1 <- indices[1:which(indices == max_index)]
  group2 <- indices[(which(indices == max_index) + 1):length(indices)]

  switched <- TRUE

  ## Run while points are switched
  while (switched) {
    switched <- FALSE

    tab1 <- m[group1, , drop = FALSE]
    tab2 <- m[group2, , drop = FALSE]

    chisq_values <- cpp_switch_docs(tab1, tab2)
    current_max <- max(chisq_values, na.rm = TRUE)

    if (current_max > max_chisq) {
      switched <- TRUE
      to_switch <- indices[which.max(chisq_values)]
      if (to_switch %in% group1) {
        group1 <- group1[group1 != to_switch]
        group2 <- c(group2, to_switch)
      } else {
        group2 <- group2[group2 != to_switch]
        group1 <- c(group1, to_switch)
      }
      max_chisq <- current_max
    }
  }

  return(list(
    indices1 = group1,
    indices2 = group2,
    chisq = max_chisq
  ))
}


#' Remove features from dtm of each group base don cc_test and tsj
#'
#' @param m global dtm
#' @param indices1 indices of documents of group 1
#' @param indices2 indices of documents of group 2
#' @param cc_test maximum contingency coefficient value for the
#' feature to be kept in both groups.
#' @param tsj minimum feature frequency in the dtm
#'
#' @details
#' Internal function, not to be used directly
#'
#' @return a list of two character vectors : `cols1` is the name of features to
#' keep in group 1, `cols2` the name of features to keep in group 2

select_features <- function(m, indices1, indices2, cc_test = 0.3, tsj = 3) {
  ## features count for each group
  tab1 <- colSums(m[indices1, , drop = FALSE])
  tab2 <- colSums(m[indices2, , drop = FALSE])
  ## Total number of features in each group
  nfeat_group1 <- sum(tab1)
  nfeat_group2 <- sum(tab2)
  ## Observed frequency of features
  observed <- rbind(tab1, tab2)
  ## Expected frequency of features
  expected_prop <- (tab1 + tab2) / sum(tab1 + tab2)
  expected <- rbind(expected_prop * nfeat_group1, expected_prop * nfeat_group2)
  ## Chi2 and contingency coefficients for each feature
  feat_chisq <- colSums((observed - expected)^2 / expected)
  ## C contingency coefficient, sqrt(khi2 / (khi2 + N))
  feat_cc <- sqrt(feat_chisq / (feat_chisq + colSums(observed)))

  ## Features selection
  cols1 <- character()
  cols2 <- character()

  for (i in seq_along(feat_cc)) {
    cc <- feat_cc[i]
    name <- names(feat_cc)[i]
    ## Keep feature if cc <= cc_test and frequency > 0
    if (cc <= cc_test && tab1[i] >= tsj) {
      cols1 <- c(cols1, name)
    }
    if (cc <= cc_test && tab2[i] >= tsj) {
      cols2 <- c(cols2, name)
    }
    ## If cc > cc_test, only keep feature in the group
    ## where observed frequency > expected frequency
    if (cc > cc_test) {
      if (tab1[i] > expected[1, i] && tab1[i] >= tsj) {
        cols1 <- c(cols1, name)
      }
      if (tab2[i] > expected[2, i] && tab2[i] >= tsj) {
        cols2 <- c(cols2, name)
      }
    }
  }

  return(list(
    cols1 = cols1,
    cols2 = cols2
  ))
}


#' Split a dtm into two clusters with reinert algorithm
#'
#' @param dtm to be split, passed by `rainette`
#' @param cc_test maximum contingency coefficient value for the
#' feature to be kept in both groups.
#' @param tsj minimum feature frequency in the dtm
#'
#' @details
#' Internal function, not to be used directly
#'
#' @return
#' An object of class `hclust` and `rainette`

cluster_tab <- function(dtm, cc_test = 0.3, tsj = 3) {
  ## Remove documents and features with zero occurrences
  dtm <- dtm[rowSums(dtm) > 0, colSums(dtm) > 0]
  ## Remove features with zero
  m <- quanteda::convert(dtm, to = "matrix")
  storage.mode(m) <- "integer"

  ## First step : CA partition

  indices <- order_docs(m)
  res <- cpp_split_tab(m, indices)
  max_index <- res$max_index
  max_chisq <- res$max_chisq

  ## Second step : switching docs

  res <- switch_docs(m, indices, max_index, max_chisq)
  indices1 <- res$indices1
  indices2 <- res$indices2
  chisq <- res$chisq

  ## Third step : features selection

  res <- select_features(m, indices1, indices2, cc_test, tsj)
  cols1 <- res$cols1
  cols2 <- res$cols2

  return(list(
    groups = list(
      docvars(dtm, "rainette_uc_id")[indices1],
      docvars(dtm, "rainette_uc_id")[indices2]
    ),
    tabs = list(
      dtm[indices1, cols1],
      dtm[indices2, cols2]
    ),
    height = chisq
  ))
}

library(quanteda)
context("simple reinert classification")

m <- matrix(
  c(
    1L, 1L, 1L, 0L, 0L,
    0L, 0L, 1L, 1L, 0L,
    1L, 0L, 0L, 0L, 0L,
    1L, 1L, 1L, 1L, 0L,
    1L, 1L, 1L, 1L, 1L,
    1L, 0L, 0L, 0L, 0L,
    1L, 1L, 1L, 0L, 0L
  ),
  ncol = 5,
  byrow = TRUE,
)
mini_dfm <- as.dfm(m)


## docs_order_by_ca

indices <- rainette:::order_docs(mini_dfm)

test_that("order_docs", {
  skip_if_not_installed("quanteda.textmodels")
  if (requireNamespace("quanteda.textmodels", quietly = TRUE)) {
    ca_values <- quanteda.textmodels::textmodel_ca(mini_dfm)$rowcoord[, 1]
    expect_equal(order(ca_values), indices)
  }
})

## split_tab_by_chisq

tab <- convert(mini_dfm, to = "matrix")
storage.mode(tab) <- "integer"
res_split <- rainette:::cpp_split_tab(m, indices)

test_that("split_tab", {
  manual_res <- data.frame(index = -1, chisq = -1)
  for (i in 1:6) {
    tab1 <- m[indices[1:i], , drop = FALSE]
    tab2 <- m[indices[(i + 1):7], , drop = FALSE]
    expect_warning(chisq <- chisq.test(rbind(colSums(tab1), colSums(tab2)))$statistic)
    manual_res <- rbind(manual_res, c(index = indices[i], chisq = chisq))
  }
  max <- which.max(manual_res$chisq)
  expect_equal(res_split$max_index, manual_res$index[max])
  expect_equal(res_split$max_chisq, manual_res$chisq[max], tolerance = 1e-06)
})

## switch_docs

test_that("switch_docs", {
  ## Wrong max_index et max_chisq
  max_chisq <- 4.5
  max_index <- 1
  res_switch <- rainette:::switch_docs(tab, indices, max_index, max_chisq)
  expect_equal(res_switch$indices1, c(3, 6))
  expect_equal(res_switch$indices2, c(7, 4, 5, 2, 1))
  expect_equal(res_switch$chisq, res_split$max_chisq)
})

## select_features

test_that("select_features", {
  res_switch <- rainette:::switch_docs(tab, indices, res_split$max_index, res_split$max_chisq)
  indices1 <- res_switch$indices1
  indices2 <- res_switch$indices2
  res <- rainette:::select_features(tab, indices1, indices2)
  expect_equal(res$cols1, character(0))
  expect_equal(res$cols2, c("feat2", "feat3", "feat4"))
  indices1 <- c(4, 3, 5, 6, 2)
  indices2 <- c(7, 1)
  res <- rainette:::select_features(tab, indices1, indices2)
  expect_equal(res$cols1, c("feat1", "feat3", "feat4"))
  expect_equal(res$cols2, character(0))
})

## rainette

test_that("rainette on mini_dfm", {
  res <- rainette(mini_dfm, k = 2, min_segment_size = 1, min_split_members = 3)
  expect_is(res, c("rainette", "hclust"))
  expect_equal(res$group, c(2, 2, 1, 2, 2, 1, 2))
  expect_equal(which(res$group == 1), indices[1:which(indices == res_split$max_index)])
  expect_equal(res$height, res_split$max_chisq)
})


mini_corpus <- head(data_corpus_inaugural, n = 2)
mini_corpus <- split_segments(mini_corpus)
dtm <- dfm(tokens(mini_corpus, remove_punct = TRUE), tolower = TRUE)
dtm <- dfm_remove(dtm, stopwords("en"))
dtm <- dfm_wordstem(dtm, language = "english")
dtm <- dfm_trim(dtm, min_termfreq = 3)

test_that("Stopping if tab too small", {
  expect_message(
    res <- rainette(dtm, k = 12, min_segment_size = 10, min_split_members = 3),
    "! Tab to be splitted is not big enough. Stopping after iteration"
  )
  expect_equal(max(res$group), 9)
})

test_that("Stopping if no more group > min_split_members", {
  expect_message(
    res <- rainette(dtm, k = 3, min_segment_size = 10, min_split_members = 12),
    "^! No more group bigger than min_split_members. Stopping after iteration"
  )
  expect_equal(max(res$group), 2)
  expect_message(
    res <- rainette(dtm, k = 3, min_segment_size = 10, min_split_members = 15),
    "! No computed clusters. Returning NULL."
  )
  expect_null(res)
})

test_that("all uce belong to the same group", {
  res <- rainette(dtm, k = 3, min_segment_size = 5)
  dtm_uc <- rainette:::merge_segments(dtm, min_segment_size = 5)
  expect_equal(anyDuplicated(unique(cbind(docvars(dtm_uc, "rainette_uc_id"), res$group))), 0)
  expect_equal(anyDuplicated(unique(cbind(docvars(dtm_uc, "rainette_uce_id"), res$group))), 0)
})

corpus <- head(data_corpus_inaugural, n = 10)
corpus <- split_segments(corpus)
dtm <- dfm(tokens(corpus, remove_punct = TRUE), tolower = TRUE)
dtm <- dfm_remove(dtm, stopwords("en"))
dtm <- dfm_trim(dtm, min_termfreq = 3)

test_that("groups correspondance is ok", {
  res <- rainette(dtm, k = 6, min_segment_size = 5)
  df <- data.frame(res$uce_groups)
  tab <- table(df[, 4], df[, 5])
  expect_equal(sum(tab > 0), 6)
  tab <- table(df[, 3], df[, 4])
  expect_equal(sum(tab > 0), 5)
})


## Special cases

test_that("dfm with 'document' feature is ok", {
  m_document <- m
  colnames(m_document) <- c("document", "éé", "ùù", "feat4", "feat5")
  dfm_document <- as.dfm(m_document)
  res_document <- rainette(dfm_document, k = 2, min_segment_size = 1, min_split_members = 3)
  res <- rainette(mini_dfm, k = 2, min_segment_size = 1, min_split_members = 3)
  expect_equal(res$group, res_document$group)
  expect_equal(res$height, res_document$height)
})


## min_members deprecation

test_that("min_members is deprecated", {
  expect_warning(
    res_deprecated <- rainette(mini_dfm, k = 2, min_segment_size = 1, min_members = 3),
    "min_members is deprecated. Use min_split_members instead."
  )
  res_deprecated$call <- NULL
  res <- rainette(mini_dfm, k = 2, min_segment_size = 1, min_split_members = 3)
  res$call <- NULL
  expect_equal(
    res,
    res_deprecated
  )
})

## Compute signed Khi2 statistic from different sizes

compute_chi2 <- function(n_both, n1, n2, n_tot) {
  tab <- matrix(
    c(n_both, n1 - n_both, n2 - n_both, n_tot - n1 - n2 + n_both),
    nrow = 2
  )
  suppressWarnings(cs <- stats::chisq.test(tab))
  res <- cs$statistic
  if (cs$expected[1, 1] > n_both) {
    res <- -res
  }
  return(res)
}

## Compute data frame of group membership at each level for a clustering

get_groups <- function(res) {
  groups <- purrr::imap(
    res$uce_groups,
    function(group, level) {
      paste(level, group, sep = ".")
    }
  )
  names(groups) <- seq_along(groups)
  dplyr::bind_cols(groups)
}

## Compute size and chi2 for all combinations of two groups from two
## clustering results

groups_crosstab <- function(groups1, groups2, min_members, min_chi2) {
  # Frequencies of each group in first clustering
  g1_count <- groups1 %>%
    tidyr::pivot_longer(
      everything(),
      names_to = "level1",
      values_to = "g1"
    ) %>%
    dplyr::count(.data$level1, .data$g1, name = "n1")

  # Frequencies of each group in second clustering
  g2_count <- groups2 %>%
    tidyr::pivot_longer(
      everything(),
      names_to = "level2",
      values_to = "g2"
    ) %>%
    dplyr::count(.data$level2, .data$g2, name = "n2")

  # Compute frequencies of all groups combinations
  res <- purrr::map_dfr(names(groups1), function(level1) {
    purrr::map_dfr(names(groups2), function(level2) {
      df <- dplyr::tibble(
        g1 = groups1[[level1]],
        g2 = groups2[[level2]]
      )
      df %>%
        dplyr::count(.data$g1, .data$g2) %>%
        tidyr::complete(.data$g1, .data$g2, fill = list(n = 0)) %>%
        dplyr::rename(n_both = n) %>%
        dplyr::mutate(level1 = .env$level1, level2 = .env$level2)
    })
  })

  # Add group count, chi-squared statistic and members
  res %>%
    # Filter on members
    dplyr::filter(.data$n_both > .env$min_members) %>%
    dplyr::left_join(g1_count, by = c("level1", "g1")) %>%
    dplyr::left_join(g2_count, by = c("level2", "g2")) %>%
    # Compute chi-squared
    dplyr::rowwise() %>%
    dplyr::mutate(
      chi2 = compute_chi2(.data$n_both, .data$n1, .data$n2, nrow(groups1)) %>%
        unname()
    ) %>%
    dplyr::ungroup() %>%
    # Filter chi-squared
    dplyr::filter(.data$chi2 > .env$min_chi2) %>%
    dplyr::mutate(interclass = paste(.data$g1, .data$g2, sep = "x"))
}

# Add members list to each crossing group

crosstab_add_members <- function(tab, groups1, groups2) {
  tab %>%
    rowwise() %>%
    dplyr::mutate(
      members = list(
        which(
          groups1[[.data$level1]] == .data$g1 &
            groups2[[.data$level2]] == .data$g2
        )
      )
    ) %>%
    dplyr::ungroup() %>%
    # Filter groups with same members
    dplyr::distinct(.data$members, .keep_all = TRUE) %>%
    dplyr::mutate(id = 1:n())
}

# Only keep most associated pairs of crossing group

crosstab_keep_max <- function(tab) {
  # We only keep crossing group where the second member is the group with
  # the highest chi2 association with the first, and the first member is the
  # group with the highest chi2 association with the second
  res <- tab %>%
    dplyr::group_by(.data$g1) %>%
    dplyr::mutate(max_g2 = .data$g2[which.max(.data$chi2)]) %>%
    dplyr::group_by(.data$g2) %>%
    dplyr::mutate(max_g1 = .data$g1[which.max(.data$chi2)]) %>%
    dplyr::filter(.data$max_g1 == .data$g1 & .data$max_g2 == .data$g2) %>%
    dplyr::ungroup()

  res
}

## Compute size of each pairs of intersection classes. Lower triangle
## and diagonal at 1 not to be selected as a partition afterward.

cross_sizes <- function(crosstab) {
  sizes <- matrix(1, nrow = nrow(crosstab), ncol = nrow(crosstab))
  for (i in 1:(nrow(crosstab) - 1)) {
    for (j in (i + 1):nrow(crosstab)) {
      sizes[i, j] <- length(
        intersect(crosstab$members[[i]], crosstab$members[[j]])
      )
    }
  }
  sizes
}

## Compute next level partitions with for loop and progress bar

next_partitions_for <- function(partitions, sizes) {
  res <- list()

  progressr::with_progress({
    p <- progressr::progressor(steps = length(partitions) / 100)

    ## for each previous partition
    for (i in seq_along(partitions)) {
      partition <- partitions[[i]]
      size_inter <- colSums(sizes[partition, ])
      classes_ok <- which(size_inter == 0)
      res_current <- vector(mode = "list", length = length(classes_ok))
      for (j in seq_along(classes_ok)) {
        res_current[[j]] <- c(partition, classes_ok[j])
      }
      res[[i]] <- res_current
      if (i %% 100 == 0) p()
    }
  })
  res <- unlist(res, recursive = FALSE)

  if (length(res) == 0) {
    return(NULL)
  }

  res
}

## Compute next level partitions with mclapply (doesn't work on Windows)

next_partitions_parallel <- function(partitions, sizes) {
  ## for each previous partition
  res <- parallel::mclapply(partitions, function(partition) {
    size_inter <- colSums(sizes[partition, ])
    classes_ok <- which(size_inter == 0)
    res_current <- vector(mode = "list", length = length(classes_ok))
    for (j in seq_along(classes_ok)) {
      res_current[[j]] <- c(partition, classes_ok[j])
    }
    res_current
  })
  res <- unlist(res, recursive = FALSE)

  if (length(res) == 0) {
    return(NULL)
  }

  res
}

## From computed partitions, select the optimal ones and add group
## membership

get_optimal_partitions <- function(partitions, cross_groups, n_tot, full) {
  ## Compute group memberships from a vector of groups
  compute_members <- function(cluster) {
    cluster <- unlist(cluster)
    groups <- rep(NA, n_tot)
    for (i in seq_along(cluster)) {
      members <- unlist(cross_groups$members[cluster[i]])
      groups[members] <- i
    }
    groups
  }

  progressr::with_progress({
    p <- progressr::progressor(steps = length(partitions) + 1)

    ## Compute data frame of results
    res <- purrr::imap_dfr(partitions, function(partition, k) {
      if (is.null(partition)) {
        return(NULL)
      }
      dim(partition) <- NULL
      out <- dplyr::tibble(clusters = partition, k = k + 1) %>%
        dplyr::rowwise() %>%
        ## Compute size and sum of Khi2 for each partition
        dplyr::mutate(
          chi2 = sum(cross_groups$chi2[.data$clusters]),
          n = sum(cross_groups$n_both[.data$clusters])
        )
      ## If full computation, keep both max chi2 and max size
      if (full) {
        out <- out %>%
          ## Filter partitions with max size or max chi2 for each k
          dplyr::group_by(.data$k) %>%
          dplyr::filter(
            .data$n == max(.data$n) | .data$chi2 == max(.data$chi2)
          ) %>%
          ## If several partitions with same n, keep max chi2
          dplyr::group_by(.data$k, .data$n) %>%
          dplyr::slice_max(.data$chi2) %>%
          ## If several partitions with same chi2, keep max n
          dplyr::group_by(.data$k, .data$chi2) %>%
          dplyr::slice_max(.data$n) %>%
          dplyr::ungroup()
      } else {
        ## If not full computation, only keep max chi2
        out <- out %>%
          ## Filter partitions with max chi2 for each k
          dplyr::group_by(.data$k) %>%
          dplyr::slice_max(.data$chi2) %>%
          ## If several partitions with same chi2, keep max n
          dplyr::group_by(.data$k, .data$chi2) %>%
          dplyr::slice_max(.data$n) %>%
          dplyr::ungroup()
      }
      p()
      out
    })
    res <- res %>%
      dplyr::rowwise() %>%
      # Add group membership for each clustering
      dplyr::mutate(groups = list(compute_members(.data$clusters))) %>%
      # Replace cross groups ids by their name
      dplyr::mutate(
        clusters = list(cross_groups$interclass[.data$clusters])
      ) %>%
      dplyr::ungroup()
  })
  res
}


#' Corpus clustering based on the Reinert method - Double clustering
#'
#' @param x either a quanteda dfm object or the result of [rainette()]
#' @param y if `x` is a [rainette()] result, this must be another [rainette()]
#'   result from same dfm but with different uc size.
#' @param max_k maximum number of clusters to compute
#' @param min_segment_size1 if `x` is a dfm, minimum uc size for first clustering
#' @param min_segment_size2 if `x` is a dfm, minimum uc size for second clustering
#' @param full if TRUE, all crossed groups are kept to compute optimal partitions, otherwise
#'   only the most mutually associated groups are kept.
#' @param doc_id character name of a dtm docvar which identifies source documents.
#' @param min_members minimum members of each cluster
#' @param min_chi2 minimum chi2 for each cluster
#' @param parallel if TRUE, use `parallel::mclapply` to compute partitions
#'   (won't work on Windows, uses more RAM)
#' @param uc_size1 deprecated, use min_segment_size1 instead
#' @param uc_size2 deprecated, use min_segment_size2 instead
#' @param ... if `x` is a dfm object, parameters passed to [rainette()] for both
#'   simple clusterings
#'
#' @details
#' You can pass a quanteda dfm as `x` object, the function then performs two simple
#' clustering with varying minimum uc size, and then proceed to find optimal partitions
#' based on the results of both clusterings.
#'
#' If both clusterings have already been computed, you can pass them as `x` and `y` arguments
#' and the function will only look for optimal partitions.
#'
#' `doc_id` must be provided unless the corpus comes from `split_segments`,
#' in this case `segment_source` is used by default.
#'
#' If `full = FALSE`, computation may be much faster, but the chi2 criterion will be the only
#' one available for best partition detection, and the result may not be optimal.
#'
#' For more details on optimal partitions search algorithm, please see package vignettes.
#'
#' @return
#' A tibble with optimal partitions found for each available value of `k` as rows, and the following
#' columns :
#'
#' - `clusters` list of the crossed original clusters used in the partition
#' - `k` the number of clusters
#' - `chi2` sum of the chi2 value of each cluster
#' - `n` sum of the size of each cluster
#' - `groups` group membership of each document for this partition (`NA` if not assigned)
#'
#' @seealso [rainette()], [cutree_rainette2()], [rainette2_plot()], [rainette2_explor()]
#'
#' @references
#'
#' - Reinert M, Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte, Cahiers de l'analyse des données, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
#' - Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Méthodologie Sociologique, Volume 26, Numéro 1, 1990.  \doi{10.1177/075910639002600103}
#'
#' @export
#'
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 3)
#'
#' res1 <- rainette(dtm, k = 5, min_segment_size = 10)
#' res2 <- rainette(dtm, k = 5, min_segment_size = 15)
#'
#' res <- rainette2(res1, res2, max_k = 4)
#' }
#'
rainette2 <- function(
  x,
  y = NULL,
  max_k = 5,
  min_segment_size1 = 10,
  min_segment_size2 = 15,
  doc_id = NULL,
  min_members = 10,
  min_chi2 = 3.84,
  parallel = FALSE,
  full = TRUE,
  uc_size1,
  uc_size2,
  ...
) {
  ## Check for deprecated uc_size1 argument
  if (!missing(uc_size1)) {
    warning("! uc_size1 is deprecated. Use min_segment_size1 instead.")
    if (missing(min_segment_size1)) {
      min_segment_size1 <- uc_size1
    }
  }
  ## Check for deprecated uc_size2 argument
  if (!missing(uc_size2)) {
    warning("! uc_size2 is deprecated. Use min_segment_size2 instead.")
    if (missing(min_segment_size2)) {
      min_segment_size2 <- uc_size2
    }
  }

  ## If passed a dfm, compute both clustering
  if (inherits(x, "dfm")) {
    dtm <- x
    message(
      "  Computing first clustering with min_segment_size1 = ",
      min_segment_size1
    )
    x <- rainette::rainette(
      dtm,
      k = max_k,
      min_segment_size = min_segment_size1,
      doc_id = doc_id,
      min_split_members = min_members,
      ...
    )
    message(
      "  Computing second clustering with min_segment_size2 = ",
      min_segment_size2
    )
    y <- rainette::rainette(
      dtm,
      k = max_k,
      min_segment_size = min_segment_size2,
      doc_id = doc_id,
      min_split_members = min_members,
      ...
    )
  }

  ## max_k should not be higher than the smallest k in both clustering
  max_k_res <- min(max(x$group, na.rm = TRUE), max(y$group, na.rm = TRUE))
  if (max_k_res < max_k) {
    message("! Lowering max_k from ", max_k, " to ", max_k_res)
    max_k <- max_k_res
  }

  message("  Searching for best partitions...")
  message("  Computing size 2 partitions...")

  progressr::with_progress({
    p <- progressr::progressor(steps = 5)

    ## Compute data frame of groups at each k for both clusterings
    groups1 <- get_groups(x)
    groups2 <- get_groups(y)
    p()
    ## Check if both clusterings have same ndoc
    if (nrow(groups1) != nrow(groups2)) {
      stop("! Number of documents in both clustering results must be the same")
    }
    ## Total number of documents
    n_tot <- nrow(groups1)

    ## Compute sizes and chi2 of every crossing between classes
    ## of both clusterings (intersection classes)
    cross_groups <- groups_crosstab(groups1, groups2, min_members, min_chi2)
    p()
    ## Add members list to each crossing group
    cross_groups <- crosstab_add_members(cross_groups, groups1, groups2)
    p()

    if (nrow(cross_groups) < 2) {
      stop(
        "! Not enough valid classes to continue. You may try a lower min_members value."
      )
    }
    if (!full) {
      cross_groups <- crosstab_keep_max(cross_groups)
    }

    ## Matrix of number of common elements between crossing groups
    sizes <- cross_sizes(cross_groups)
    p()

    ## Compute size 2 partitions
    partitions <- list()
    ## Size 2 partitions : pairs of cross groups with no common elements
    partitions[[1]] <- which(sizes == 0, arr.ind = TRUE, useNames = FALSE) %>%
      asplit(1)
    p()
  })

  ## Partitions computing method
  next_partitions <- ifelse(
    parallel,
    next_partitions_parallel,
    next_partitions_for
  )

  ## Compute partitions of size > 2
  if (max_k > 2) {
    for (k in 1:(max_k - 2)) {
      message("  Computing size ", k + 2, " partitions...")
      next_part <- next_partitions(partitions[[k]], sizes)
      if (is.null(next_part)) {
        message("! No more partitions found, stopping at k=", k + 1)
        break
      }
      partitions[[k + 1]] <- next_part
    }
  }

  message("  Selecting best partitions...")
  ## Select optimal partitions and add group membership for each one
  res <- get_optimal_partitions(partitions, cross_groups, n_tot, full)

  message("  Done.")

  class(res) <- c("rainette2", class(res))
  attr(res, "full") <- full

  res
}

library(quanteda)
context("double reinert classification")

mini_corpus <- head(data_corpus_inaugural, n = 2)
mini_corpus <- split_segments(mini_corpus, 5)
dtm <- dfm(tokens(mini_corpus, remove_punct = TRUE), tolower = TRUE)
dtm <- dfm_remove(dtm, stopwords("en"))
dtm <- dfm_wordstem(dtm, language = "english")
dtm <- dfm_trim(dtm, min_termfreq = 3)

res1 <- rainette(dtm, k = 5, min_segment_size = 2, min_split_members = 3)
res2 <- rainette(dtm, k = 5, min_segment_size = 3, min_split_members = 3)
res12 <- rainette2(dtm, max_k = 5, min_segment_size1 = 2, min_segment_size2 = 3, min_members = 3)

res <- rainette2(res1, res2, min_members = 2)
res_notfull <- rainette2(res1, res2, min_members = 2, full = FALSE)

test_that("compute_chi2", {
  tab <- matrix(c(45, 121 - 45, 257 - 45, 1213 - 121 - 257 + 45), nrow = 2)
  expect_equal(rainette:::compute_chi2(45, 121, 257, 1213),
    chisq.test(tab)$statistic)
  tab <- matrix(c(0, 12, 25, 121 - 12 - 25), nrow = 2)
  expect_equal(rainette:::compute_chi2(0, 12, 25, 121),
    suppressWarnings(-chisq.test(tab)$statistic))
})

test_that("get_groups", {
  expect_equal(dim(rainette:::get_groups(res1)), c(316, 4))
  expect_equal(substr(rainette:::get_groups(res1)[[2]],3,3), as.character(res1$uce_groups[[2]]))
  expect_equal(substr(rainette:::get_groups(res1)[[2]],1,2), rep("2.", length(res1$group)))
})

test_that("groups_crosstab", {
  g1 <- tibble(c(1,1,2,2), c(1,1,2,3))
  g2 <- tibble(c(1,2,1,2), c(3,1,2,1))
  colnames(g1) <- 1:2
  colnames(g2) <- 1:2
  n_tot <- 4
  tab <- rainette:::groups_crosstab(g1, g2, min_members = -Inf, min_chi2 = -Inf)

  expect_equal(nrow(tab), 25)
  expect_equal(
    colnames(tab),
    c("g1", "g2", "n_both", "level1", "level2", "n1", "n2", "chi2", "interclass")
  )
  tmp <- tab %>% dplyr::filter(level1 == 2, level2 == 2, g1 == 2, g2 == 2)
  expect_equal(tmp$n_both, 1)
  expect_equal(tmp$n1, 1)
  expect_equal(tmp$n2, 1)
  expect_equal(tmp$chi2, unname(rainette:::compute_chi2(tmp$n_both, tmp$n1, tmp$n2, n_tot)))
  expect_equal(tmp$interclass, "2x2")
})

test_that("crosstab_add_members and filtering", {
  g1 <- tibble(c(1,1,1,1,2), c(1,1,2,3,3))
  g2 <- tibble(c(1,1,1,1,2), c(3,1,3,1,2))
  colnames(g1) <- 1:2
  colnames(g2) <- 1:2
  n_tot <- 5
  tab <- rainette:::groups_crosstab(g1, g2, min_members = 2, min_chi2 = 0.5)
  tab <- rainette:::crosstab_add_members(tab, g1, g2)

  expect_equal(nrow(tab), 1)
  expect_equal(tab$g1, 1)
  expect_equal(tab$g2, 1)
  expect_equal(tab$n_both, 4)
  expect_equal(tab$interclass, "1x1")
  expect_equal(tab$chi2,
    suppressWarnings(unname(chisq.test(matrix(c(4,0,0,1), nrow=2))$statistic)))
  expect_equal(tab$members, list(1:4))
})

test_that("crosstab_keep_max", {
  df <- tribble(
    ~ g1, ~ g2, ~chi2,
      11,   21,   10,
      11,   22,   5,
      11,   23,   8,
      12,   21,   9, 
      12,   22,   8,
      12,   23,   10,
      13,   21,   8,
      13,   22,   2,
      13,   23,   2
  )
  res <- crosstab_keep_max(df)
  expect_equal(res$g1, c(11, 12))
  expect_equal(res$g2, c(21, 23))
})

test_that("cross_sizes", {
  tab <- tibble(interclass = c("1.1x1.1", "1.2x1.1", "2.1x2.2", "2.1x1.1"),
    members = list(c(1,2,3), c(4,5,6), c(1,2), 5), id = 1:4)
  sizes <- rainette:::cross_sizes(tab)
  expect_equal(sizes, structure(c(1, 1, 1, 1, 0, 1, 1, 1, 2, 0, 1, 1, 0, 1, 0, 1), .Dim = c(4L,
    4L)))
})

test_that("next_partitions", {
  tab <- tibble(interclass = c("1.1x1.1", "1.2x1.1", "2.1x2.2", "2.1x1.1"),
    members = list(c(1,2,3), c(4,5,6), c(1,2), 7), id = 1:4)
  sizes <- rainette:::cross_sizes(tab)
  partitions <- list(tab$id)
  partitions[[2]] <- which(sizes == 0, arr.ind = TRUE, useNames = FALSE) %>%
      asplit(1)

  partitions[[3]] <- rainette:::next_partitions_for(partitions[[2]], sizes)
  expect_equal(partitions[[3]], list(c(1, 2, 4), c(2, 3, 4)))
  partitions[[3]] <- rainette:::next_partitions_parallel(partitions[[2]], sizes)
  expect_equal(partitions[[3]], list(c(1, 2, 4), c(2, 3, 4)))
  expect_equal(rainette:::next_partitions_for(partitions[[3]], sizes), NULL)
  expect_equal(rainette:::next_partitions_parallel(partitions[[3]], sizes), NULL)
})


test_that("get_optimal_partitions", {
  valid <- tibble(interclass = c("1x1", "2x2", "3x3", "4x4"),
                  n_both = c(3, 3, 2, 1),
                  chi2 = c(4, 5, 10, 4),
                  members = list(c(1,2,3), c(4,5,6), c(1,2), 7),
                  id = 1:4)
  partitions <- list(
    list(c(1, 2), c(2, 3), c(1, 4)),
    list(c(1, 2, 4)))
  n_tot <- 7

  # With full = TRUE
  res <- rainette:::get_optimal_partitions(partitions, valid, n_tot, full = TRUE)
  tmp <- res[res$k == 2,]
  expect_equal(tmp$clusters, list(c("1x1", "2x2"), c("2x2", "3x3")))
  expect_equal(tmp$chi2, c(9, 15))
  expect_equal(tmp$n, c(6, 5))
  expect_equal(tmp$groups, list(c(1,1,1,2,2,2,NA),c(2,2,NA,1,1,1,NA)))
  tmp <- res[res$k == 3,]
  expect_equal(tmp$clusters, list(c("1x1", "2x2", "4x4")))
  expect_equal(tmp$chi2, 13)
  expect_equal(tmp$n, 7)
  expect_equal(tmp$groups, list(c(1,1,1,2,2,2,3)))

  # With full = FALSE
  res <- rainette:::get_optimal_partitions(partitions, valid, n_tot, full = FALSE)
  tmp <- res[res$k == 2,]
  expect_equal(tmp$clusters, list(c("2x2", "3x3")))
  expect_equal(tmp$chi2, 15)
  expect_equal(tmp$n, 5)
  expect_equal(tmp$groups, list(c(2,2,NA,1,1,1,NA)))
  tmp <- res[res$k == 3,]
  expect_equal(tmp$clusters, list(c("1x1", "2x2", "4x4")))
  expect_equal(tmp$chi2, 13)
  expect_equal(tmp$n, 7)
  expect_equal(tmp$groups, list(c(1,1,1,2,2,2,3)))
})

test_that("rainette2 gives the same result on dtm and on two clustering results", {
  expect_equal(res[,c("k", "chi2", "n")], res12[,c("k", "chi2", "n")])
  expect_equal(res$groups, res12$groups)
  expect_equal(res$clusters, res12$clusters)
})

test_that("rainette2 display message when stopping before max_k", {
  res1 <- rainette(dtm, k = 4, min_segment_size = 2, min_split_members = 30)
  res2 <- rainette(dtm, k = 4, min_segment_size = 3, min_split_members = 30)
  expect_message(res <- rainette::rainette2(res1, res2, max_k = 4, min_members = 50),
    "^! No more partitions found, stopping at k=2")
  expect_equal(max(res$k), 2)
})

test_that("rainette2_plot error if full=FALSE and criterion=n", {
  expect_error(
    rainette2_plot(res_notfull, dtm, k = 2, criterion = "n"),
    "if rainette2 has been computed with full=FALSE, only 'chi2' criterion is available"
  )
  expect_error(
    rainette2_plot(res, dtm, k = 2, criterion = "n"),
    NA
  )
})

test_that("same results with different values of full and parallel", {
  res_f <- rainette2(res1, res2, max_k = 5, min_members = 2, full = TRUE)
  res_fp <- rainette2(res1, res2, max_k = 5, min_members = 2, full = TRUE, parallel = TRUE)
  expect_equal(res_f, res_fp)
  expect_equal(attr(res_f, "full"), TRUE)
  expect_equal(attr(res_fp, "full"), TRUE)

  res_c <- rainette2(res1, res2, max_k = 5, min_members = 2, full = FALSE)
  res_cp <- rainette2(res1, res2, max_k = 5, min_members = 2, full = FALSE, parallel = TRUE)
  expect_equal(res_c, res_cp)
  expect_equal(attr(res_c, "full"), FALSE)
  expect_equal(attr(res_cp, "full"), FALSE)
})

test_that("plot functions class checking", {
  expect_error(rainette2_plot(res1, k = 5), "res must be a rainette2 result object")
  expect_error(rainette_plot(res12, k = 5), "res must be a rainette result object")
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_chisq <- function(T1, T2, Tsum) {
  .Call('_rainette_cpp_chisq', PACKAGE = 'rainette', T1, T2, Tsum)
}

cpp_split_tab <- function(tab, indices) {
  .Call('_rainette_cpp_split_tab', PACKAGE = 'rainette', tab, indices)
}

cpp_switch_docs <- function(tab1, tab2) {
  .Call('_rainette_cpp_switch_docs', PACKAGE = 'rainette', tab1, tab2)
}

library(quanteda)
context("Rcpp functions")

m <- matrix(
  c(1L,1L,1L,0L,0L,
    0L,0L,1L,1L,0L,
    1L,0L,0L,0L,0L,
    1L,1L,1L,1L,0L,
    1L,1L,1L,1L,1L,
    1L,0L,0L,0L,0L,
    1L,1L,1L,0L,0L
    ),
  ncol = 5,
  byrow = TRUE,
)
mini_dfm <- as.dfm(m)

# cpp_chisq

test_that("cpp_chisq", {
  set.seed(13370)
  t1 <- as.integer(round(runif(100, 5, 50)))
  t2 <- as.integer(round(runif(100, 10, 200)))
  expect_equal(rainette:::cpp_chisq(t1, t2, t1 + t2),
               unname(chisq.test(cbind(t1, t2))$statistic))
})

#' Cut a tree into groups
#'
#' @param tree the hclust tree object to be cut
#' @param ... arguments passed to other methods
#'
#' @return
#' A vector with group membership.
#'
#' @details
#' If `tree` is of class `rainette`, invokes [cutree_rainette()]. Otherwise, just run [stats::cutree()].
#'
#' @export

cutree <- function(tree, ...) {
  if (inherits(tree, "rainette")) {
    return(cutree_rainette(tree, ...))
  }
  if (inherits(tree, "rainette2")) {
    return(cutree_rainette2(tree, ...))
  }
  stats::cutree(tree, ...)
}


#' Cut a rainette result tree into groups of documents
#'
#' @param hres the `rainette` result object to be cut
#' @param k the desired number of clusters
#' @param h unsupported
#' @param ... arguments passed to other methods
#'
#' @return
#' A vector with group membership.
#'
#' @export

cutree_rainette <- function(hres, k = NULL, h = NULL, ...) {
  if (!is.null(h)) {
    stop("cutree_rainette only works with k argument")
  }
  hres$uce_groups[[k - 1]]
}

#' Cut a rainette2 result object into groups of documents
#'
#' @param res the `rainette2` result object to be cut
#' @param k the desired number of clusters
#' @param criterion criterion to use to choose the best partition. `chi2` means
#'    the partition with the maximum sum of chi2, `n` the partition with the
#'    maximum size.
#' @param ... arguments passed to other methods
#'
#' @return
#' A vector with group membership.
#'
#' @seealso [rainette2_complete_groups()]
#' @importFrom rlang .env
#' @export

cutree_rainette2 <- function(res, k, criterion = c("chi2", "n"), ...) {
  criterion <- match.arg(criterion)
  line <- res %>%
    dplyr::filter(k == .env$k)
  if (criterion == "chi2") {
    line <- line %>%
      dplyr::slice_max(.data$chi2)
  }
  if (criterion == "n") {
    line <- line %>%
      dplyr::slice_max(.data$n)
  }
  line %>%
    dplyr::pull(.data$groups) %>%
    unlist()
}


#' Complete groups membership with knn classification
#'
#' Starting with groups membership computed from a `rainette2` clustering,
#' every document not assigned to a cluster is reassigned using a k-nearest
#' neighbour classification.
#'
#' @param dfm dfm object used for `rainette2` clustering.
#' @param groups group membership computed by `cutree` on `rainette2` result.
#' @param k number of neighbours considered.
#' @param ... other arguments passed to `FNN::knn`.
#'
#' @return
#' Completed group membership vector.
#'
#' @seealso [cutree_rainette2()], [FNN::knn()]
#'
#' @export

rainette2_complete_groups <- function(dfm, groups, k = 1, ...) {
  if (!requireNamespace("FNN", quietly = TRUE)) {
    stop(
      "Package \"FNN\" needed for this function to work. Please install it.",
      call. = FALSE
    )
  }

  m <- quanteda::convert(dfm, to = "matrix")

  test <- m[is.na(groups), ]
  train <- m[!is.na(groups), ]
  train_groups <- groups[!is.na(groups)]

  new_groups <- FNN::knn(train, test, train_groups, k = k, ...)
  groups[is.na(groups)] <- new_groups

  groups
}

library(quanteda)
context("cutree functions")

mini_corpus <- head(data_corpus_inaugural, n = 2)
mini_corpus <- split_segments(mini_corpus)
dtm <- dfm(tokens(mini_corpus, remove_punct = TRUE), tolower = TRUE)
dtm <- dfm_remove(dtm, stopwords("en"))
dtm <- dfm_wordstem(dtm, language = "english")
dtm <- dfm_trim(dtm, min_termfreq = 3)
res <- rainette(dtm, k = 3, min_segment_size = 10)

## cutree

test_that("generic cutree still works", {
  hc <- hclust(dist(USArrests))
  expect_length(cutree(hc, h = 250), 50)
})

test_that("cutree_rainette is ok", {
  expect_error(cutree(res, h = 200))
  expect_equal(cutree(res, k = 3), c(
    3L, 3L, 3L, 3L, 3L, 3L, 3L, 2L, 2L, 2L, 2L, 3L, 3L, 2L, 2L,
    2L, 2L, 2L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L, 2L, 1L, 1L, 2L, 2L,
    2L, 2L, 2L, 2L, 2L, 2L, 3L, 3L, 1L, 1L
  ))
})

test_that("rainette2_complete_groups", {
  skip_if_not_installed("FNN")
  res1 <- rainette(dtm, k = 5, min_segment_size = 2, min_split_members = 3)
  res2 <- rainette(dtm, k = 5, min_segment_size = 3, min_split_members = 3)
  res_double <- rainette2(res1, res2, min_members = 2)

  groups <- cutree(res_double, k = 4)
  expect_equal(sum(is.na(rainette2_complete_groups(dtm, groups))), 0)
})

## Generate a "terms bar plot", based on terms keyness for a group

keyness_barplot <- function(
  tab,
  range = NULL,
  title = "",
  title_color = "firebrick3",
  stat_col = "chi2",
  n_terms = NULL,
  text_size = 10,
  top_margin = 0,
  keyness_plot_xlab = NULL
) {
  if (is.null(keyness_plot_xlab)) {
    keyness_plot_xlab <- stat_col
  }

  ## Column with statistic values
  if (!is.null(range)) {
    stat_max <- max(range)
  } else {
    stat_max <- max(tab[[stat_col]], na.rm = TRUE)
  }
  ## Plot
  g <- ggplot(
    data = tab,
    aes(
      x = stats::reorder(.data$feature, .data[[stat_col]]),
      y = abs(.data[[stat_col]])
    )
  ) +
    geom_col(aes(fill = .data$sign), color = "white", width = 1) +
    geom_text(
      y = stat_max / 15,
      aes(label = stats::reorder(.data$feature, .data[[stat_col]])),
      hjust = 0,
      size = text_size / 2.5
    ) +
    coord_flip() +
    scale_fill_manual(
      "",
      values = c("positive" = "#a1d8ff", "negative" = "#ff7d7e")
    ) +
    guides(fill = "none") +
    labs(title = title, x = "") +
    theme_minimal() +
    theme(
      plot.title = element_text(
        size = text_size,
        face = "bold",
        hjust = 0.5,
        colour = title_color
      ),
      axis.title.x = element_text(size = text_size * 0.8),
      plot.margin = grid::unit(c(top_margin, 0.05, 0, 0), "npc"),
      panel.grid.major.x = element_blank(),
      panel.grid.minor.x = element_blank(),
      panel.background = element_rect(
        fill = grDevices::rgb(.9, .9, .9, .2),
        colour = "transparent"
      )
    )
  ## Fix x limits if necessary and remove horizontal axis values
  if (!is.null(range)) {
    g <- g +
      ggplot2::scale_y_continuous(
        keyness_plot_xlab,
        limits = range,
        breaks = NULL
      )
  } else {
    g <- g + ggplot2::scale_y_continuous(keyness_plot_xlab, breaks = NULL)
  }
  ## Adjust vertical scale if necessary
  if (nrow(tab) < n_terms) {
    limits <- levels(stats::reorder(tab$feature, tab[[stat_col]]))
    limits <- c(rep("", n_terms - length(limits)), limits)
    g <- g + ggplot2::scale_x_discrete(limits = limits, breaks = NULL)
  } else {
    g <- g + ggplot2::scale_x_discrete(breaks = NULL)
  }

  ## Align title element to the left to center it with hjust
  g <- ggplot2::ggplotGrob(g)
  g$layout$l[g$layout$name == "title"] <- 1
  g
}


## Generate a "terms wordcloud plot", based on terms keyness for a group

#' @import ggwordcloud

keyness_worcloud <- function(
  tab,
  range = NULL,
  title = "",
  title_color = "firebrick3",
  stat_col = "chi2",
  max_size = 15,
  top_margin = 0
) {
  ## Plot
  g <- ggplot(data = tab) +
    # geom_hline(yintercept = 0, color = "grey70") +
    ggwordcloud::geom_text_wordcloud(
      aes(label = .data$feature, size = .data[[stat_col]]),
      color = title_color
    ) +
    labs(title = title) +
    scale_x_continuous(stat_col) +
    theme_minimal() +
    theme(
      plot.title = element_text(
        size = 12,
        face = "bold",
        hjust = 0.5,
        colour = title_color
      ),
      plot.margin = grid::unit(c(top_margin, 0.05, 0, 0), "npc"),
      panel.grid.major.x = element_blank(),
      panel.grid.minor.x = element_blank(),
      panel.background = element_rect(
        fill = grDevices::rgb(.9, .9, .9, .3),
        colour = "transparent"
      )
    )
  ## Fix x limits if necessary and remove horizontal axis values
  if (!is.null(range)) {
    g <- g + ggplot2::scale_size_area(limits = range, max_size = max_size)
  } else {
    g <- g + ggplot2::scale_size_area(max_size = max_size)
  }

  ## Align title element to the left to center it with hjust
  g <- ggplot2::ggplotGrob(g)
  g$layout$l[g$layout$name == "title"] <- 1
  g
}


## Returns a color palette or an individual group color depending on the number of groups

groups_colors <- function(k, i = NULL) {
  ## Groups colors
  ## This is the result of the following call, put manually here
  ## for compatibility with R 3.6
  ## col <- grDevices::palette.colors(n = 10, palette = "Tableau 10")
  col <- c(
    "#4E79A7",
    "#F28E2B",
    "#E15759",
    "#76B7B2",
    "#59A14F",
    "#EDC948",
    "#B07AA1",
    "#FF9DA7",
    "#9C755F",
    "#BAB0AC"
  )
  col <- rep_len(col, k)

  if (!is.null(i)) {
    return(col[i])
  } else {
    return(col)
  }
}


## Generate a list of terms plots from a list of keyness statistic tables

keyness_plots <- function(
  tabs,
  groups,
  type = "bar",
  free_scales = FALSE,
  stat_col = "chi2",
  n_terms,
  text_size,
  top_margin = 0,
  cluster_label = NULL,
  keyness_plot_xlab = NULL,
  colors = NULL
) {
  ## Frequency and proportion of each cluster
  clust_n <- table(groups)
  clust_prop <- round(clust_n / sum(clust_n) * 100, 1)
  k <- length(tabs)
  if (is.null(colors[1])) {
    colors <- purrr::map_chr(.x = 1:k, ~ groups_colors(k, .x))
  }

  ## Min and max statistics to fix x axis in terms plots
  if (stat_col == "docprop") {
    range <- c(0, 1)
  } else {
    range <- NULL
    if (!free_scales) {
      max_stat <- max(purrr::map_dbl(tabs, function(tab) {
        v <- tab[[stat_col]]
        if (length(v) == 0) {
          return(0)
        }
        max(v)
      }))
      range <- c(0, max_stat)
    }
  }

  purrr::map(1:k, function(i) {
    # in case label for class is provided use it and do not show number of class
    cluster_num <- paste0(" ", i)
    if (length(cluster_label) > 1) {
      cluster_label <- cluster_label[i]
      cluster_num <- ""
    }
    if (k <= 6) {
      label <- ifelse(is.null(cluster_label), "Cluster", cluster_label)
      title <- paste0(
        label,
        cluster_num,
        "\nn = ",
        clust_n[i],
        " (",
        clust_prop[i],
        "%)"
      )
    } else if (k <= 8) {
      label <- ifelse(is.null(cluster_label), "Cluster", cluster_label)
      title <- paste0(label, cluster_num, "\nn = ", clust_n[i])
    } else {
      label <- ifelse(is.null(cluster_label), "Cl.", cluster_label)
      title <- paste0(label, cluster_num, "\nn = ", clust_n[i])
    }
    if (type == "bar") {
      if (is.null(text_size)) {
        text_size <- 10
      }
      keyness_barplot(
        tabs[[i]],
        range,
        title = title,
        title_color = colors[i],
        stat_col = stat_col,
        n_terms,
        text_size = text_size,
        top_margin,
        keyness_plot_xlab = keyness_plot_xlab
      )
    } else {
      if (is.null(text_size)) {
        text_size <- 15
      }
      keyness_worcloud(
        tabs[[i]],
        range,
        title = title,
        title_color = colors[i],
        stat_col = stat_col,
        max_size = text_size,
        top_margin
      )
    }
  })
}

# Generate the dendrogram plot for rainette_plot()

dendrogram_plot <- function(
  res,
  k,
  groups,
  text_size,
  show_na_title,
  colors = NULL
) {
  dend <- stats::as.dendrogram(res)
  max_k <- max(res$group, na.rm = TRUE)

  if (is.null(colors[1])) {
    colors <- purrr::map_chr(.x = 1:k, ~ groups_colors(k, .x))
  }
  ## Cut the dendrogram if necessary
  if (k < max_k) {
    dend <- cut(dend, res$height[max_k - k])$upper
    ## Double conversion to "balance" the dendrogram
    dend <- stats::as.dendrogram(stats::as.hclust(dend))
    dend <- dendextend::set(dend, "labels", 1:k)
  }

  ## Style labels and branches
  dendextend::labels_colors(dend) <- colors
  dend <- dend %>%
    dendextend::color_branches(k = k, col = colors) %>%
    dendextend::set("branches_lwd", 0.4)

  ## Generate plot
  dend <- dendextend::as.ggdend(dend)
  margin <- ifelse(k >= 7, 0, 0.175 - k * 0.025)
  title_size <- ifelse(is.null(text_size), 10, text_size)
  g <- ggplot(dend, nodes = FALSE) +
    scale_y_continuous(breaks = NULL) +
    theme(
      plot.margin = grid::unit(c(0.05, margin, 0, margin), "npc"),
      plot.title = element_text(hjust = 0.5, size = title_size)
    )

  ## Add NA number and proportion as title
  if (show_na_title) {
    ## Compute number of NA
    na_n <- sum(is.na(groups))
    na_prop <- round(na_n / length(groups) * 100, 1)
    g <- g +
      ggtitle(paste0("NA : ", na_n, " (", na_prop, "%)"))
  }

  g
}


#' Generate a clustering description plot from a rainette result
#'
#' @param res result object of a `rainette` clustering
#' @param dtm the dfm object used to compute the clustering
#' @param k number of groups. If NULL, use the biggest number possible
#' @param type type of term plots : barplot or wordcloud
#' @param n_terms number of terms to display in keyness plots
#' @param free_scales if TRUE, all the keyness plots will have the same scale
#' @param measure statistics to compute
#' @param show_negative if TRUE, show negative keyness features
#' @param text_size font size for barplots, max word size for wordclouds
#' @param show_na_title if TRUE, show number of NA as plot title
#' @param cluster_label define a specific term for clusters identification in keyness plots.
#'   Default is "Cluster" or "Cl." depending on the number of groups. If a vector of length > 1,
#'   define the cluster labels manually.
#' @param keyness_plot_xlab define a specific x label for keyness plots.
#' @param colors vector of custom colors for cluster titles and branches (in the order of the clusters)
#'
#' @seealso [quanteda.textstats::textstat_keyness()], [rainette_explor()], [rainette_stats()]
#'
#' @return
#' A gtable object.
#'
#' @export
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 3)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' rainette_plot(res, dtm)
#' rainette_plot(
#'   res,
#'   dtm,
#'   cluster_label = c("Assets", "Future", "Values"),
#'   colors = c("red", "slateblue", "forestgreen")
#' )
#' }
#'
#' @import ggplot2
#' @import dendextend

rainette_plot <- function(
  res,
  dtm,
  k = NULL,
  type = c("bar", "cloud"),
  n_terms = 15,
  free_scales = FALSE,
  measure = c("chi2", "lr", "frequency", "docprop"),
  show_negative = FALSE,
  text_size = NULL,
  show_na_title = TRUE,
  cluster_label = NULL,
  keyness_plot_xlab = NULL,
  colors = NULL
) {
  if (!inherits(res, "rainette")) {
    stop("res must be a rainette result object")
  }

  type <- match.arg(type)
  measure <- match.arg(measure)
  stat_col <- stat_col(measure)
  if (type == "cloud") {
    warning(
      "wordcloud plots will soon be deprecated. Please use type = \"bar\" instead."
    )
    show_negative <- FALSE
  }

  ## Maximum number of clusters
  max_k <- max(res$group, na.rm = TRUE)

  ## Get groups
  if (is.null(k)) {
    groups <- res$group
    k <- max_k
  } else {
    if (k < 2 || k > max_k) {
      stop("k must be between 2 and ", max_k)
    }
    groups <- rainette::cutree_rainette(res, k)
  }

  if (is.null(colors[1])) {
    colors <- purrr::map_chr(.x = 1:k, ~ groups_colors(k, .x))
  }
  ## Keyness statistics
  tabs <- rainette::rainette_stats(groups, dtm, measure, n_terms, show_negative)

  ## Graph layout
  lay <- matrix(
    c(
      rep(1, k),
      rep(2:(k + 1), 2)
    ),
    nrow = 3,
    ncol = k,
    byrow = TRUE
  )
  plots <- list()

  ## Dendrogram
  plots[[1]] <- dendrogram_plot(
    res,
    k,
    groups,
    text_size,
    show_na_title,
    colors = colors
  )

  ## Add terms plots
  plots <- c(
    plots,
    keyness_plots(
      tabs,
      groups,
      type,
      free_scales,
      stat_col,
      n_terms,
      text_size,
      cluster_label = cluster_label,
      keyness_plot_xlab = keyness_plot_xlab,
      colors = colors
    )
  )

  ## Generate grid
  gridExtra::grid.arrange(grobs = plots, layout_matrix = lay)
}

## Generate barplot of cluster sizes

frequency_barplot <- function(groups, k, text_size, colors = NULL) {
  if (is.null(colors[1])) {
    colors <- purrr::map_chr(.x = 1:k, ~ groups_colors(k, .x))
  }

  # Compute cluster sizes and number of NA
  tab <- table(groups, exclude = NULL)
  # Replace NA with placeholder to avoid error in R 4.7
  names(tab)[is.na(names(tab))] <- "<rainette_NA>"
  freq <- data.frame(tab)
  n_na <- sum(is.na(groups))
  title <- paste0("Clusters size\n(NA = ", n_na, ")")
  colnames(freq) <- c("Group", "n")
  # Transform back placeholder to NA
  freq$Group[freq$Group == "<rainette_NA>"] <- NA

  # Generate barplot
  g <- ggplot(freq) +
    geom_col(aes(x = .data$Group, y = .data$n, fill = .data$Group)) +
    scale_fill_manual(values = colors, na.value = "grey20") +
    guides(fill = "none") +
    ggtitle(title) +
    theme(
      plot.title = element_text(size = 10, face = "bold", hjust = 0.5),
      plot.margin = grid::unit(c(0.05, 0.05, 0, 0), "npc"),
      axis.title.x = element_text(size = text_size * 0.8),
      axis.title.y = element_text(size = text_size * 0.8)
    )
}

#' Generate a clustering description plot from a rainette2 result
#'
#' @param res result object of a `rainette2` clustering
#' @param dtm the dfm object used to compute the clustering
#' @param k number of groups. If NULL, use the biggest number possible
#' @param criterion criterion to use to choose the best partition. `chi2` means
#'    the partition with the maximum sum of chi2, `n` the partition with the
#'    maximum size.
#' @param complete_groups if TRUE, documents with NA cluster are reaffected by
#'    k-means clustering initialised with current groups centers.
#' @param type type of term plots : barplot or wordcloud
#' @param n_terms number of terms to display in keyness plots
#' @param free_scales if TRUE, all the keyness plots will have the same scale
#' @param measure statistics to compute
#' @param show_negative if TRUE, show negative keyness features
#' @param text_size font size for barplots, max word size for wordclouds
#'
#' @return
#' A gtable object.
#'
#' @seealso [quanteda.textstats::textstat_keyness()], [rainette2_explor()], [rainette2_complete_groups()]
#'
#' @export
#'
#' @import ggplot2

rainette2_plot <- function(
  res,
  dtm,
  k = NULL,
  criterion = c("chi2", "n"),
  complete_groups = FALSE,
  type = c("bar", "cloud"),
  n_terms = 15,
  free_scales = FALSE,
  measure = c("chi2", "lr", "frequency", "docprop"),
  show_negative = FALSE,
  text_size = 10
) {
  if (!inherits(res, "rainette2")) {
    stop("res must be a rainette2 result object")
  }

  type <- match.arg(type)
  measure <- match.arg(measure)
  criterion <- match.arg(criterion)
  stat_col <- stat_col(measure)

  if (type == "cloud") {
    warning(
      "wordcloud plots will soon be deprecated. Please use type = \"bar\" instead."
    )
    show_negative <- FALSE
  }

  ## Stop if not full and criterion == "n"
  if (
    (is.null(attr(res, "full")) || !attr(res, "full")) && criterion != "chi2"
  ) {
    stop(
      "if rainette2 has been computed with full=FALSE, only 'chi2' criterion is available"
    )
  }

  ## Maximum number of clusters
  max_k <- max(res$k, na.rm = TRUE)

  ## Get groups
  if (is.null(k) || k < 2 || k > max_k) {
    stop("k must be between 2 and ", max_k)
  }
  groups <- rainette::cutree_rainette2(res, k, criterion)
  if (complete_groups) {
    groups <- rainette::rainette2_complete_groups(dtm, groups)
  }

  ## Keyness statistics
  tabs <- rainette::rainette_stats(groups, dtm, measure, n_terms, show_negative)

  plots <- list()

  ## Barplot of clusters size
  plots[[1]] <- frequency_barplot(groups, k, text_size)

  ## Add terms plots
  plots <- c(
    plots,
    keyness_plots(
      tabs,
      groups,
      type,
      free_scales,
      stat_col,
      n_terms,
      text_size,
      top_margin = 0.05
    )
  )

  ## Graph layout
  nrow <- ifelse(k > 5, 2, 1)
  index <- 1:(k + 1)
  if (k > 5 && k %% 2 == 0) {
    index <- c(index, NA)
  }
  lay <- matrix(index, nrow = nrow, byrow = TRUE)

  ## Generate grid
  gridExtra::grid.arrange(grobs = plots, layout_matrix = lay)
}

#' Generate cluster keyness statistics from a rainette result
#'
#' @param groups groups membership computed by `cutree_rainette` or `cutree_rainette2`
#' @param dtm the dfm object used to compute the clustering
#' @param measure statistics to compute
#' @param n_terms number of terms to display in keyness plots
#' @param show_negative if TRUE, show negative keyness features
#' @param max_p maximum keyness statistic p-value
#'
#' @seealso [quanteda.textstats::textstat_keyness()], [rainette_explor()], [rainette_plot()]
#'
#' @return
#' A list with, for each group, a data.frame of keyness statistics for the most specific
#' n_terms features.
#'
#' @export
#' @examples
#' \donttest{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 3)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' groups <- cutree_rainette(res, k = 3)
#' rainette_stats(groups, dtm)
#' }

rainette_stats <- function(
  groups,
  dtm,
  measure = c("chi2", "lr", "frequency", "docprop"),
  n_terms = 15,
  show_negative = TRUE,
  max_p = 0.05
) {
  measure <- match.arg(measure)
  stat_col <- stat_col(measure)

  groups_list <- sort(unique(groups))
  groups_list <- groups_list[!is.na(groups_list)]
  tabs <- purrr::map(groups_list, function(group) {
    select <- (groups == group & !is.na(groups))

    empty_tab <- tibble(
      feature = character(0),
      col = double(0),
      sign = character(0)
    )
    names(empty_tab) <- c("feature", stat_col(measure), "sign")

    ## Keyness
    if (measure %in% c("chi2", "lr")) {
      tab <- quanteda.textstats::textstat_keyness(
        dtm,
        select,
        measure = measure
      ) %>%
        dplyr::as_tibble() %>%
        dplyr::arrange(desc(abs(.data[[stat_col]])))
      if (all(is.nan(tab$p))) {
        return(empty_tab)
      }
      if (show_negative) {
        tab <- tab %>%
          dplyr::filter(p <= max_p) %>%
          dplyr::slice(1:.env$n_terms) %>%
          dplyr::mutate(
            sign = ifelse(.data[[stat_col]] > 0, "positive", "negative"),
            sign = factor(sign, levels = c("positive", "negative"))
          )
      } else {
        tab <- tab %>%
          dplyr::filter(.data[[stat_col]] > 0, p <= max_p) %>%
          dplyr::slice(1:.env$n_terms) %>%
          dplyr::mutate(sign = "positive")
      }
    }
    if (measure %in% c("frequency", "docprop")) {
      tmp_dtm <- quanteda::dfm_subset(dtm, select)
      if (all(colSums(tmp_dtm) == 0)) {
        return(empty_tab)
      }
      tab <- quanteda.textstats::textstat_frequency(tmp_dtm) %>%
        dplyr::as_tibble() %>%
        dplyr::mutate(docprop = .data$docfreq / ndoc(tmp_dtm)) %>%
        dplyr::arrange(desc(.data[[stat_col]])) %>%
        dplyr::slice(1:.env$n_terms) %>%
        dplyr::mutate(sign = "positive")
    }
    return(tab)
  })

  tabs
}

#' Shiny gadget for rainette clustering exploration
#'
#' @param res result object of a `rainette` clustering
#' @param dtm the dfm object used to compute the clustering
#' @param corpus_src the quanteda corpus object used to compute the dtm
#'
#' @seealso `rainette_plot`
#'
#' @return
#' No return value, called for side effects.
#'
#' @examples
#' \dontrun{
#' require(quanteda)
#' corpus <- data_corpus_inaugural
#' corpus <- head(corpus, n = 10)
#' corpus <- split_segments(corpus)
#' tok <- tokens(corpus, remove_punct = TRUE)
#' tok <- tokens_remove(tok, stopwords("en"))
#' dtm <- dfm(tok, tolower = TRUE)
#' dtm <- dfm_trim(dtm, min_docfreq = 3)
#' res <- rainette(dtm, k = 3, min_segment_size = 15)
#' rainette_explor(res, dtm, corpus)
#' }
#'
#' @export
#'
#' @import shiny
#' @import miniUI

rainette_explor <- function(res, dtm = NULL, corpus_src = NULL) {
  ## Check for res, dtm and corpus_src values and consistency
  if (is.null(dtm)) {
    stop(
      "rainette_explor must be called with a result object and its associated dtm."
    )
  }
  if (!is.null(corpus_src) && ndoc(corpus_src) != ndoc(dtm)) {
    stop("corpus_src and dtm must have the same number of documents.")
  }

  ## If res is a rainette2 result, launch rainette2_explor
  if (inherits(res, "rainette2")) {
    stop(
      "trying to use rainette_explor() on a rainette2 result object. Use rainette2_explor() instead."
    )
  }

  if (length(res$group) != ndoc(dtm)) {
    stop("res and dtm must have the same number of documents.")
  }

  res_name <- deparse(substitute(res))
  dtm_name <- deparse(substitute(dtm))
  max_n_groups <- max(res$group, na.rm = TRUE)

  ui <- miniPage(
    gadgetTitleBar("Clustering exploration", left = NULL),
    tags$head(tags$style(rainette_explor_css())),
    miniTabstripPanel(
      miniTabPanel(
        "Summary",
        icon = shiny::icon("chart-bar"),
        miniContentPanel(
          fillRow(
            flex = c(1, 3),
            fillCol(
              flex = c(10, 1),
              id = "side",
              div(
                sliderInput(
                  "k",
                  label = "Number of clusters",
                  value = max_n_groups,
                  min = 2,
                  max = max_n_groups,
                  step = 1
                ),
                selectInput(
                  "measure",
                  "Statistics",
                  choices = c(
                    "Keyness - Chi-squared" = "chi2",
                    "Keyness - Likelihood ratio" = "lr",
                    "Frequency - Terms" = "frequency",
                    "Frequency - Documents proportion" = "docprop"
                  )
                ),
                numericInput(
                  "n_terms",
                  label = "Number of terms to display",
                  value = 20,
                  min = 5,
                  max = 30,
                  step = 1
                ),
                conditionalPanel(
                  "input.measure != 'docprop'",
                  checkboxInput(
                    "same_scales",
                    label = "Force same scales",
                    value = TRUE
                  )
                ),
                checkboxInput(
                  "show_negative",
                  label = "Show negative values",
                  value = FALSE
                ),
                sliderInput(
                  "text_size",
                  label = "Text size",
                  value = 12,
                  min = 6,
                  max = 20,
                  step = 1
                )
              ),
              actionButton(
                "get_r_code",
                class = "btn-success",
                icon = icon("code"),
                label = gettext("Get R code")
              )
            ),
            fillCol(
              id = "main",
              plotOutput("rainette_plot", height = "100%")
            )
          )
        )
      ),
      miniTabPanel(
        "Cluster documents",
        icon = shiny::icon("file-alt"),
        miniContentPanel(
          docs_sample_ui("rainette1", res)
        )
      )
    )
  )

  server <- function(input, output, session) {
    plot_code <- reactive({
      code <- paste0(
        "rainette_plot(\n  ",
        res_name,
        ", ",
        dtm_name,
        ", k = ",
        input$k,
        ",\n  n_terms = ",
        input$n_terms,
        ",\n  free_scales = ",
        !input$same_scales,
        ",\n  measure = \"",
        input$measure,
        "\"",
        ",\n  show_negative = ",
        input$show_negative,
        ifelse(
          input$text_size != "10",
          paste0(",\n  text_size = ", input$text_size),
          ""
        ),
        "\n)"
      )
      code
    })

    cutree_code <- reactive({
      paste0(
        "cutree_rainette(",
        res_name,
        ", k = ",
        input$k,
        ")"
      )
    })

    generate_code <- reactive({
      code <- "## Clustering description plot\n"
      code <- paste0(code, plot_code())
      code <- paste0(code, "\n## Groups\n")
      code <- paste0(code, cutree_code())
      code
    })

    output$rainette_plot <- renderPlot({
      eval(parse(text = plot_code()), envir = .GlobalEnv)
    })

    ## Code export modal dialog
    observeEvent(input$get_r_code, {
      code <- generate_code()
      showModal(modalDialog(
        title = gettext("Export R code"),
        size = "l",
        HTML(paste0(
          "Code to generate the current plot and compute groups :",
          "<pre><code>",
          paste(highr::hi_html(code), collapse = "\n"),
          "</code></pre>"
        )),
        easyClose = TRUE
      ))
    })

    current_k <- shiny::reactive(input$k)
    docs_sample_server("rainette1", res, corpus_src, current_k)

    # Handle the Done button being pressed.
    observeEvent(input$done, {
      ## Generate code
      code <- generate_code()
      out <- paste0(
        "\n-------- Start exported code --------\n\n",
        paste(code, collapse = "\n"),
        "\n\n--------- End exported code ---------\n"
      )
      cat(out)
      shiny::stopApp()
    })
  }

  shiny::runGadget(
    ui,
    server
  )
}

---
title: "[fr] Description des algorithmes"
author: "Julien Barnier"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{[fr] Description des algorithmes}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


L'objectif de ce document est de décrire les algorithmes utilisés pour les classifications dans `rainette` et leur implémentation. À noter que cette implémentation s'est appuyée sur deux éléments principaux :

- les trois articles de Max Reinert cités dans les [références]
- le code source d'Iramuteq, notamment les fichiers `CHD.R` et `chdtxt.R`. Le code a cependant été quasiment entièrement réécrit, et certaines portions ont été implémentées en C++ via [Rcpp](https://www.rcpp.org/).

## Algorithme de classification simple

### Matrice de départ

La classification simple est une classification descendante hiérarchique (CDH).

Le tableau de départ est la matrice termes-documents croisant les documents du corpus et les termes qui les composent. Cette matrice est une matrice binaire de présence/absence du terme dans le document, et non une matrice d'occurrences.

On a donc un tableau de ce type :

```{r echo=FALSE}
set.seed(1337)
m <- matrix(sample(0:1, 20, replace = TRUE), nrow = 4)
rownames(m) <- paste0("doc", 1:4)
colnames(m) <- c("partir", "un", "jour", "sans", "retour")
m
```

À noter que si les documents sont des segments obtenus avec `split_segments()`, les segments trop courts d'un même document peuvent être regroupés entre eux au moment de la classification pour atteindre la taille minimale indiquée par l'argument `min_segment_size` de `rainette()`.

### Division maximisant le Khi2

On souhaite diviser cette matrice termes-documents en deux groupes de documents aussi "différents" que possible. La méthode Reinert consiste à sélectionner le regroupement qui maximise la statistique du χ² du tableau regroupé.

Par exemple, dans le tableau ci-dessus, si on regroupe `doc1` avec `doc2` et `doc3` avec `doc4`, on obtient le tableau suivant :

```{r echo=FALSE}
tmp <- rbind(colSums(m[1:2,]), colSums(m[3:4,]))
rownames(tmp) <- c("doc1 + doc2", "doc3 + doc4")
tmp
```

On peut calculer la valeur du χ² de ce tableau. Cette statistique est un indicateur de la "distance" entre les deux groupes de documents en ce qui concerne la distribution des termes : plus la valeur du χ² du tableau regroupé est élevée et plus les deux groupes sont différents, plus elle est faible et plus ils se ressemblent.

Dans un cas aussi simple, on peut effectuer tous les regroupements de documents possibles et déterminer lequel correspond au χ² maximal, mais avec des données réelles la complexité augmente trop rapidement.

On opère donc de la manière suivante :

- on effectue une analyse factorielle des correspondances de la matrice termes-documents, et on ordonne les documents selon leur coordonnée sur le premier axe de cette AFC.
- on regroupe tour à tour les documents entre eux selon cet ordonnancement : d'abord le point avec la coordonnée la plus basse *vs* tous les autres, puis les deux points avec les coordonnées les plus basses *vs* tous les autres, etc. On calcule à chaque fois le χ² correspondant et on conserve au final le regroupement qui le maximise.
- à partir de ce regroupement, on effectue une réaffectation des documents : on change tour à tour chaque document de classe, et on regarde si cela fait augmenter le χ². Si c'est le cas on conserve cette nouvelle affectation.
- on recommence cette opération de réaffectation jusqu'à ce qu'elle ne permette plus d'augmenter la valeur du χ².

On obtient donc au final deux groupes de documents et deux matrices termes-documents correspondantes.

### Sélection des termes

L'étape suivante consiste en une sélection des termes dans chacun des deux tableaux pour les prochaines itérations :

- on regarde la fréquence de chaque terme, et on supprime les termes qui apparaissent moins de 3 fois (cette valeur est modifiable via l'argument `tsj` de `rainette()`).
- on compare également l'effectif observé de chaque terme avec l'effectif attendu sous l'hypothèse d'indépendance de la répartition du terme entre les deux groupes, et on en déduit un coefficient de contingence pour ce terme. Si ce coefficient de contingence est supérieur à 0.3, on ne conserve le terme que dans le tableau dans lequel il est surreprésenté (le seuil de 0.3 est modifiable via l'argument `cc_test` de `rainette()`).


### Classification descendante hiérarchique

Les étapes précédentes permettent de scinder un corpus en deux groupes. Pour obtenir une hiérarchie de classes, on commence par scinder le corpus entier, puis on recommence de la même manière avec le groupe obtenu comportant le plus de documents (sauf si le tableau en question est trop petit pour calculer une AFC, ou si son effectif est inférieur à l'argument `min_split_members` de `rainette()`).

En procédant de cette manière `k - 1` fois, on obtient une classification descendante hiérarchique (CDH) en `k` groupes, et le dendrogramme correspondant (la hauteur des branches du dendrogramme étant la valeur maximale du χ² trouvée au moment du split).

![](img/dendro5.png)



## Algorithme de classification double

La classification double s'applique à un corpus découpé en segments. L'objectif est d'obtenir des classes plus robustes en croisant les résultats de deux classifications simples calculées avec des tailles minimales de segments différentes.

La classification double consiste donc à effectuer deux classifications simples avec des valeurs distinctes de `min_segment_size`. On obtient donc deux dendrogrammes différents qui déterminent des classes ici numérotées de 1 à 6 :

![](img/double_cdh1.png)

### Calcul des classes croisées

La première étape consiste à croiser les classes des deux CDH :

![](img/double_cdh_full.png)

On prend chaque classe de la première CDH, et on la croise avec chaque classe de la seconde (y compris si les classes ne sont pas de même niveau `k`). On appelle la nouvelle classe obtenue une "classe croisée" :

- pour chaque classe croisée, on calcule l'effectif du croisement, c'est à dire le nombre de documents présents dans les deux classes.
- à partir de cet effectif, de l'effectif de chaque classe et du nombre total de documents, on peut créer un tableau croisé d'appartenance aux deux classes : on calcule alors le χ² de ce tableau, qui représente une mesure de "l'association" entre les deux classes.
- si deux classes croisées différentes comportent les mêmes documents, on ne conserve que l'une d'entre elles.
- si une classe croisée comporte un effectif inférieur à la valeur de l'argument `min_members` de `rainette2()`, elle n'est pas conservée.
- si une classe croisée comporte un χ² d'association inférieur à la valeur de l'argument `min_chi2` de `rainette2()`, elle n'est pas conservée.

<table>
  <thead>
    <tr>
      <th align="center">Classe croisée</th>
      <th align="right">Effectif</th>
      <th align="right">χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>1</span></span></td><td align="right">25</td><td align="right">41.2</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>2</span></span></td><td align="right">102</td><td align="right">30.1</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>3</span></span></td><td align="right">59</td><td align="right">87.3</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>5</span></span></td><td align="right">0</td><td align="right">-</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>6</span></span></td><td align="right">13</td><td align="right">6.2</td></tr>
    <tr><td align="center"><span class="cross"><span>2</span><span>1</span></span></td><td align="right">0</td><td align="right">-</td></tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>6</span></span></td><td align="right">5</td><td align="right">3.1</td></tr>

  </tbody>
</table>

Une fois ces classes croisées définies, deux type d'analyse sont possibles.

#### Analyse "full"

La première, dite analyse "full" (argument `full = TRUE` de `rainette2()`), conserve pour la suite toutes les classes croisées dont l'effectif est supérieur à zéro (si celles-ci n'ont pas déjà été filtrées à l'étape précédente). Dans l'exemple ci-dessus, on conserverait donc :

<table>
  <thead>
    <tr>
      <th align="center">Classe croisée</th>
      <th align="right">Effectif</th>
      <th align="right">χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>1</span></span></td><td align="right">25</td><td align="right">41.2</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>2</span></span></td><td align="right">102</td><td align="right">30.1</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>3</span></span></td><td align="right">59</td><td align="right">87.3</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>6</span></span></td><td align="right">13</td><td align="right">6.2</td></tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>6</span></span></td><td align="right">5</td><td align="right">3.1</td></tr>

  </tbody>
</table>

#### Analyse "classique"

La seconde, dite analyse "classique" (argument `full = FALSE` de `rainette2()`) ne conserve une classes croisée que si les deux groupes croisés sont mutuellement les plus associés. Ainsi, si on considère la classe croisée issue du croisement du groupe *x* et du groupe *y*, on ne conserve cette classe que si *y* est le groupe ayant le χ² d'association maximal avec *x*, et si en même temps *x* est le groupe ayant le χ² d'association maximal avec *y*.

Dans ce cas le nombre de classes croisées considérées est beaucoup plus réduit :

<table>
  <thead>
    <tr>
      <th align="center">Classe croisée</th>
      <th align="right">Effectif</th>
      <th align="right">χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>3</span><span>3</span></span></td><td align="right">18</td><td align="right">70.1</td></tr>
    <tr><td align="center"><span class="cross"><span>4</span><span>6</span></span></td><td align="right">21</td><td align="right">58.2</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>5</span></span></td><td align="right">25</td><td align="right">61.0</td></tr>
  </tbody>
</table>


### Calcul et sélection des partitions optimales

L'étape suivante consiste à déterminer, à partir des classes croisées calculées précédemment, toutes les partitions possibles de nos documents en 2, 3, 4... classes croisées. Plus précisément, on essaie de déterminer tous les groupes de 2, 3, 4... classes croisées qui n'ont aucun élément en commun.

On commence par déterminer les partitions de taille 2, c'est-à-dire tous les ensembles de deux classes croisées n'ayant aucun élément commun. On calcule pour chaque partition son effectif total et la somme de ses χ² d'association.

<table>
  <thead>
    <tr>
      <th align="center">Partition</th>
      <th align="right">Effectif total</th>
      <th align="right">Somme des χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>1</span></span>
        <span class="cross"><span>2</span><span>3</span></span>
      </td>
      <td align="right">47</td>
      <td align="right">62.3</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>4</span></span>
        <span class="cross"><span>3</span><span>5</span></span>
      </td>
      <td align="right">36</td>
      <td align="right">51.0</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>2</span><span>4</span></span>
        <span class="cross"><span>1</span><span>6</span></span>
      </td>
      <td align="right">29</td>
      <td align="right">38.2</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">143</td>
      <td align="right">68.7</td>
    </tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
  </tbody>
</table>

On sélectionne parmi ces partitions celles qui sont considérées comme les "meilleures", c'est-à-dire :

- si on fait une analyse "full" (`full = TRUE`), on sélectionne celle ayant la somme de χ² d'association la plus forte, et celle ayant l'effectif total le plus élevé.
- si on fait une analyse "classique" (`full = FALSE`), on sélectionne uniquement celle ayant la somme de χ² d'association la plus forte (on ne peut pas utiliser l'effectif total comme critère de sélection dans ce cas-là).

<table>
  <thead>
    <tr>
      <th align="center">Partition</th>
      <th align="right">Effectif total</th>
      <th align="right">Somme des χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">143</td>
      <td align="right">68.7</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span>
        <span class="cross"><span>2</span><span>6</span></span>
      </td>
      <td align="right">86</td>
      <td align="right">98.0</td>
    </tr>
  </tbody>
</table>

On fait de même pour les partitions de taille 3 : on détermine les différentes partitions possibles et on conserve les meilleures.

<table>
  <thead>
    <tr>
      <th align="center" style="width: 50%;">Partition</th>
      <th align="right">Effectif total</th>
      <th align="right">Somme des χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
        <span class="cross"><span>2</span><span>5</span></span>
      </td>
      <td align="right">189</td>
      <td align="right">91.3</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span>
        <span class="cross"><span>2</span><span>6</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">113</td>
      <td align="right">108.1</td>
    </tr>
  </tbody>
</table>

Et on continue pour les partitions de 4 classes croisées, etc.

<table style="margin-top: 2em;">
  <thead>
    <tr>
      <th align="center" style="width: 50%;">Partition</th>
      <th align="right">Effectif total</th>
      <th align="right">Somme des χ² d'association</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span><span class="cross"><span>2</span><span>6</span></span><span class="cross"><span>5</span><span>3</span></span><span class="cross"><span>1</span><span>5</span></span>
      </td>
      <td align="right">223</td>
      <td align="right">114.7</td>
    </tr>
  </tbody>
</table>

On répète l'opération jusqu'à atteindre la valeur de l'argument `max_k` de `rainette2()`, ou bien lorsqu'il n'y a aucune partition possible au niveau considéré.

### Résultat

On obtient donc au final, pour chaque valeur de `k` de 2 à `max_k`, une sélection de partitions de classes croisées considérées comme "optimales", soit selon le critère de la somme de leur χ² d'association, soit selon celui de leur effectif total. Ces classes croisées constituent de nouveaux groupes potentiellement plus "robustes" que ceux mis en évidence par les deux CDH simples.

À noter que cette opération peut faire qu'un grand nombre de documents n'appartiennent à aucun des nouveaux groupes. `rainette` propose de réaffecter ces documents aux nouveaux groupes via une méthode rapide de type *k-nearest neighbours*, mais ceci n'est pas forcément conseillé dans la mesure où on perdrait la "robustesse" qui est justement l'objectif de la double classification.


## Différences avec Iramuteq

### Classification simple

L'implémentation de l'algorithme dans `rainette` se différencie de celle d'Iramuteq surtout dans la gestion du paramètre `min_split_members` (le nombre minimal de documents dans une classe).

Dans `rainette`, `min_split_members` est uniquement utilisé au moment du choix du tableau suivant à splitter : si aucun tableau n'a un effectif supérieur à ce paramètre, l'algorithme s'arrête même si on n'a pas atteint la valeur souhaitée de `k`.

Dans Iramuteq, l'algorithme se répète `k` fois dans tous les cas, et il procède à la fin à un regroupement des classes dont l'effectif est inférieur à l'effectif minimal souhaité. Ce regroupement s'effectue en fusionnant ces classes en "remontant" le dendrogramme.

L'avantage de l'implémentation dans `rainette` est qu'on ne "casse" pas la logique du dendrogramme, et qu'on obtient donc comme résultat un arbre complet, qu'on peut couper à la hauteur souhaitée : on peut donc explorer la classification en 2, 3 ... `k` groupes. L'inconvénient est qu'on n'a pas d'assurance de n'avoir aucune classe avec un effectif inférieur à `min_split_members` : on doit procéder à des regroupements manuels si nécessaire.

### Classification double

La méthode proposée par Iramuteq correspond plutôt à la méthode "full" de ce document : après avoir déterminé les classes croisées, on conserve toutes celles ayant un effectif et une valeur de χ² d'association suffisants.

L'algorithme reste très proche de celui d'Iramuteq jusqu'à l'étape de la sélection de partition optimale. Ici on procède à une recherche exhaustive des partitions à partir des classes croisées sélectionnées (ce qui peut occasionner des calculs potentiellement très longs si `min_members` est faible et `max_k` élevé), tandis qu'Iramuteq procède autrement.

Au niveau des résultats, Iramuteq ne renvoie qu'une seule partition qu'il juge optimale, tandis que `rainette` retourne pour chaque valeur de `k` les meilleures partitions selon les critères de la taille ou du χ² maximum.


## Différences avec la "méthode Reinert"

Le mode de détermination des partitions optimales ne nous a pas semblé très détaillé dans les articles cités en [références], il ne nous est donc pas vraiment possible de comparer avec l'implémentation de `rainette`. A priori la méthode présentée par Max Reinert correspond à la méthode dite "classique" présentée dans ce document : après le calcul des classes croisées on ne conserve que celles dont les deux groupes croisés sont les plus associés mutuellement.

Une différence importante réside dans le fait que dans les articles cités, une fois la partition optimale déterminée, celle-ci est utilisée comme point de départ pour une affectation des documents aux classes via une méthode de type "nuées dynamiques". Ceci permet notamment de réaffecter des points qui ne feraient pas partie des classes croisées de la partition retenue.

Cette réaffectation par centre mobile n'est pas implémentée dans `rainette` (et elle ne semble pas l'être non plus dans Iramuteq). Il faut donc être vigilant au nombre et à la proportion de documents non affectés (valeur de classe à `NA`) à l'issue d'une classification double, car celui-ci peut être élevé. Une fonction `rainette2_complete_groups()` est disponible pour rattacher les points non affectés à une des classes via une méthode du type _k-nearest neighbours_, mais son utilisation n'est pas forcément conseillée.



## Références

- Reinert M., "Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte", *Cahiers de l'analyse des données*, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
- Reinert M., "Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval", *Bulletin de Méthodologie Sociologique*, Volume 26, Numéro 1, 1990. <https://doi.org/10.1177/075910639002600103>
- Reinert M., "Une méthode de classification des énoncés d’un corpus présentée à l’aide d’une application", *Les cahiers de l’analyse des données*, Tome 15, Numéro 1, 1990. <https://www.numdam.org/item/?id=CAD_1990__15_1_21_0>

---
title: "[en] Algorithms description"
author: "Julien Barnier"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{[en] Algorithms description}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The goal of this document is to describe the algorithms used in the `rainette` package and their implementation. This implementation has been based upon two main resources :

- the three papers from Max Reinert cited in [references] (in french)
- the source code of [Iramuteq](https://pratinaud.gitpages.huma-num.fr/iramuteq-website/), especially the `CHD.R` and `chdtxt.R` files. Note however that the code has been almost entirely rewritten (in part in C++ via [Rcpp](https://www.rcpp.org/)).


## Simple clustering algorithm

### Document-term matrix

The simple clustering method is a divisive hierarchical clustering applied to a document-term matrix. This matrix is binary weighted, so only the absence or presence of a term in a document is taken into account.

Here is a sample matrix :

```{r echo=FALSE}
set.seed(13)
m <- matrix(sample(0:1, 20, replace = TRUE), nrow = 4)
rownames(m) <- paste0("doc", 1:4)
colnames(m) <- c("return", "of", "the", "obra", "dinn")
m
```

Note that if documents are segments computed by `split_segments()`, segments from the same source may be merged together before clustering computation if some of them are too short and don't reach the minimum size given by the `min_segment_size` argument of `rainette()`.

### Maximal χ² splitting

In the newt step, the goal is to split the document-term matrix into two groups of documents such as these are as "different" as possible. For the Reinert method, the "best" split is the one which maximizes the χ² value of the grouped array.

For example, using the previous matrix, if we group together `doc1` with `doc2` and `doc3` with `doc4`, we get the following grouped array :

```{r echo=FALSE}
tmp <- rbind(colSums(m[1:2,]), colSums(m[3:4,]))
rownames(tmp) <- c("doc1 + doc2", "doc3 + doc4")
tmp
```

We can compute the χ² statistics of this array, which can be seen has an association coefficient between the two document groups regarding their term distributions : if the χ² is high, the two groups are different, if it is low they can be considered as similar.

In such a simple case, we could compute every possible groupings and determine which one corresponds to the maximal χ² value, but with real data the complexity and needed computations rise very rapidly.

The Reinert method therefore proceeds the following way :

- we first compute a correspondance analysis on the document-term matrix, and we sort the documents along their coordinate on the first axis.
- we group documents together successively following this ordering : first the lowest coordinate document *vs* all the others, then the two lowest ones *vs* all the others, and so on. Each time we compute the χ² value of the grouped matrix, and we keep the grouping which corresponds to the highest χ² value.
- based on this grouping, we succesively assign each document to the other group, and see if this makes the χ² rise. If it is the case, the document is kept in the other group. This operation is repeated until no reassignment makes the χ² higher.

At the end we get two groups of documents and two corresponding document-term matrices.

### Terms selection

The following step is to filter out some terms from each matrices before the next iteration :

- we compute each term frequency in both matrices, and remove terms with a frequency lower than 3 (this value can be changed using the `tsj` argument of `rainette()`).
- we also compare the observed frequency of each term and their expected frequency under independance hypothesis of distribution between the two groups. This allows to compute a contingency coefficient : if this coefficient is higher than 0.3, the term is only kept in the matrix where it is overrepresented (the 0.3 threshold can be changed using the `cc_test` argument of `rainette()`).

### Divisive hierarchical clustering

The previous steps allow to split a documents corpus into two groups. To get a hierarchy of groups, we first split the whole initial corpus, then continue by splitting the biggest resulting cluster, *ie* the one with the most documents (unless the cluster document-term matrix is too small to compute a correspondance analysis, or if its size is smaller than the value of the `min_split_members` argument of `rainette()`).

By repeating this `k - 1` times, we get a divisive hierarchical clustering of `k` clusters, and the corresponding dendrogram (the height of each dendrogram branch is given by the χ² value of the corresponding split).

![](img/dendro5.png)


## Double clustering algorithm

The double clustering only applies to a corpus which has been split into segments. The goal is to make clusters more robust by crossing the results of two single clusterings computed with different minimal segment sizes.

Doing a double clustering implies therefore implies that we first do two simple clusterings with distinct `min_segment_size` values. We then get two different dendrograms determining different clusters, here numbered from 1 to 6 :

![](img/double_cdh1.png)

### Cross-clusters computation

The first step of the algorithm is to cross all clusters from both dendrograms together :

![](img/double_cdh_full.png)

Each cluster of the first simple clustering is crossed with each cluster of the second one, even if both cluster do not belong to the same `k` level. The new resulting cluster is called a "crossed cluster" :

- for each crossed cluster, we compute its size, *ie* the number of segments belonging to both clusters.
- using this size, the size of both clusters and the total number of segments, we can create a cross table of clusters membership : we compute the χ² value of this table, which can be seen as an "association coefficient" between the two clusters.
- if two different crossed clusters hold exactly the same segments, we only keep one.
- if a crossed cluster size is lower than the value of the `min_members` argument of `rainette2()`, it is discarded.
- if a crossed cluster has an association χ² lower than the value of the `min_chi2` argument of `rainette2()`, it is discarded.

<table>
  <thead>
    <tr>
      <th align="center">Crossed cluster</th>
      <th align="right">Size</th>
      <th align="right">Association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>1</span></span></td><td align="right">25</td><td align="right">41.2</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>2</span></span></td><td align="right">102</td><td align="right">30.1</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>3</span></span></td><td align="right">59</td><td align="right">87.3</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>5</span></span></td><td align="right">0</td><td align="right">-</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>6</span></span></td><td align="right">13</td><td align="right">6.2</td></tr>
    <tr><td align="center"><span class="cross"><span>2</span><span>1</span></span></td><td align="right">0</td><td align="right">-</td></tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>6</span></span></td><td align="right">5</td><td align="right">3.1</td></tr>

  </tbody>
</table>

Once these crossed clusters have been computed, two types of analysis are possible.

#### "full" analysis

The first type, called a "full" analysis (`full = TRUE` argument of `rainette2()`), retains for the following steps all the crossed clusters with a size greater than zero (if these have not already been filtered out at the previous step). In the previous example, we would keep :

<table>
  <thead>
    <tr>
      <th align="center">Crossed cluster</th>
      <th align="right">Size</th>
      <th align="right">Association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>1</span></span></td><td align="right">25</td><td align="right">41.2</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>2</span></span></td><td align="right">102</td><td align="right">30.1</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>3</span></span></td><td align="right">59</td><td align="right">87.3</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>1</span><span>6</span></span></td><td align="right">13</td><td align="right">6.2</td></tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>6</span></span></td><td align="right">5</td><td align="right">3.1</td></tr>

  </tbody>
</table>

#### "classical" analysis

The second type, called a "classical" analysis (`full = FALSE` argument of `rainette2()`), only keeps for the following steps the crossed clusters whose clusters are mutually most associated. Thus, if we consider the crossed cluster crossing cluster *x* and cluster *y*, we only keep this crossed cluster if *y* is the cluster with the highest association χ² with *x*, and if at the same time *x* is the cluster with the highest association χ² with *y*.

In this case the number of crossed clusters kept for the following steps is much lower :

<table>
  <thead>
    <tr>
      <th align="center">Crossed cluster</th>
      <th align="right">Size</th>
      <th align="right">Association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr><td align="center"><span class="cross"><span>1</span><span>4</span></span></td><td align="right">41</td><td align="right">94.0</td></tr>
    <tr><td align="center"><span class="cross"><span>3</span><span>3</span></span></td><td align="right">18</td><td align="right">70.1</td></tr>
    <tr><td align="center"><span class="cross"><span>4</span><span>6</span></span></td><td align="right">21</td><td align="right">58.2</td></tr>
    <tr><td align="center"><span class="cross"><span>6</span><span>5</span></span></td><td align="right">25</td><td align="right">61.0</td></tr>
  </tbody>
</table>


### Determining optimal partitions

The goal of the next step is to identify, starting from the set of crossed clusters defined previously, every possible partitions of our corpus in 2, 3, 4... crossed clusters. More precisely, we try to identify each set of 2, 3, 4... crossed clusters which have no common elements.

We start by computing the size 2 partitions, *ie* every set of 2 crossed clusters without common elements. For each partition we compute is total size and the sum of its association χ².

<table>
  <thead>
    <tr>
      <th align="center">Partition</th>
      <th align="right">Total size</th>
      <th align="right">Sum of association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>1</span></span>
        <span class="cross"><span>2</span><span>3</span></span>
      </td>
      <td align="right">47</td>
      <td align="right">62.3</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>4</span></span>
        <span class="cross"><span>3</span><span>5</span></span>
      </td>
      <td align="right">36</td>
      <td align="right">51.0</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>2</span><span>4</span></span>
        <span class="cross"><span>1</span><span>6</span></span>
      </td>
      <td align="right">29</td>
      <td align="right">38.2</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">143</td>
      <td align="right">68.7</td>
    </tr>
    <tr><td align="center">...</td><td align="right">...</td><td align="right">...</td></tr>
  </tbody>
</table>

We only keep among these partitions the ones that are considered the "best" ones :

- if we are doing a "full" analysis (`full = TRUE`), we keep the one having the highest sum of χ², and the one with the highest total size.
- if we are doing a "classical" analysis (`full = FALSE`), we only keep the one with the highest sum of χ² (we can use the highest total size criterion in this case).

<table>
  <thead>
    <tr>
      <th align="center">Partition</th>
      <th align="right">Total size</th>
      <th align="right">Sum of association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">143</td>
      <td align="right">68.7</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span>
        <span class="cross"><span>2</span><span>6</span></span>
      </td>
      <td align="right">86</td>
      <td align="right">98.0</td>
    </tr>
  </tbody>
</table>

We do the same for the size 3 partitions : we identify every set of 3 non overlapping crossed clusters, and keep the "best" ones.

<table>
  <thead>
    <tr>
      <th align="center" style="width: 50%;">Partition</th>
      <th align="right">Total size</th>
      <th align="right">Sum of association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>1</span><span>2</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
        <span class="cross"><span>2</span><span>5</span></span>
      </td>
      <td align="right">189</td>
      <td align="right">91.3</td>
    </tr>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span>
        <span class="cross"><span>2</span><span>6</span></span>
        <span class="cross"><span>5</span><span>3</span></span>
      </td>
      <td align="right">113</td>
      <td align="right">108.1</td>
    </tr>
  </tbody>
</table>

And we repeat the operation for 4 crossed clusters, etc.

<table style="margin-top: 2em;">
  <thead>
    <tr>
      <th align="center" style="width: 50%;">Partition</th>
      <th align="right">Total size</th>
      <th align="right">Sum of association χ²</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <td align="center">
        <span class="cross"><span>3</span><span>4</span></span><span class="cross"><span>2</span><span>6</span></span><span class="cross"><span>5</span><span>3</span></span><span class="cross"><span>1</span><span>5</span></span>
      </td>
      <td align="right">223</td>
      <td align="right">114.7</td>
    </tr>
  </tbody>
</table>

The operation is repeated until we reach the value of the `max_k` argument passed to `rainette2()`, or until there is no possible partition of the size `k`.


### Result

At the end we get, for each `k` value from 2 to `max_k`, a selection of the "best" crossed clusters partitions, either according to the association χ² criterion or according to the total size criterion. These crossed clusters form a new set of clusters which are potentially more "robust" than the ones computed from the two simple clusterings.

After this operation, a potentially high number of segments may not belong to any cluster anymore. `rainette` allows to reassign them to the new clusters with a fast *k-nearest neighbours* method, but this may not be recommended as we would then loose the clusters robustness acquired with the double clustering.



## Differences with the "Reinert method"

The way to determine the optimal partitions did not seem completely clear to us in the articles cited in [references], so it is not really possible to compare with the `rainette` implementation. The "classical" method described in this document seems to be close from the one suggested by Max Reinert : after computing the crossed clusters, we only keep the ones where the two crossed clusters are both the most mutually associated.

One important difference is the fact that once the best partition of crossed clusters has been determined, Max Reinert suggests to use these new groups as starting points to reassign the documents to new clusters with a k-means type method. This is not implemented in `rainette` : the `rainette2_complete_groups()` allows to reassign documents without cluster using a *k-nearest neighbours* method, but this may not be recommended if you want to keep the "robustness" of the clusters computed with the double clustering.


## References

- Reinert M., "Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte", *Cahiers de l'analyse des données*, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
- Reinert M., "Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval", *Bulletin de Méthodologie Sociologique*, Volume 26, Numéro 1, 1990. <https://doi.org/10.1177/075910639002600103>
- Reinert M., "Une méthode de classification des énoncés d’un corpus présentée à l’aide d’une application", *Les cahiers de l’analyse des données*, Tome 15, Numéro 1, 1990. <https://www.numdam.org/item/?id=CAD_1990__15_1_21_0>

---
title: '[en] Introduction to rainette'
author: "Julien Barnier"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{[en] Introduction to rainette}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Corpus preparation

For a description of the Reinert clustering method and its implementation in `rainette`, please see the [algorithms description vignette](algorithms_en.html).

### Split corpus into segments

As it doesn't take into account terms frequencies but only there presence / absence, and
as it assigns each document to only one cluster, the Reinert method must be applied
to short and "homogeneous" documents. It could be ok if you work on tweets or short
answers to a specific question, but with longer documents they must first be split
into short textual segments.

You can use the `split_segments()` function for that, and it can be applied directly on a `tm` or
`quanteda` corpus. In this article we will apply it to the sample `data_corpus_inaugural`
corpus provided by `quanteda`.

```{r warning=FALSE, message=FALSE}
library(quanteda)
library(rainette)

## Split documents into segments
corpus <- split_segments(data_corpus_inaugural, segment_size = 40)
```

`split_segments` will split the original texts into smaller chunks, attempting to
respect sentences and punctuation when possible. The function takes two
arguments :

- `segment_size` : the preferred segment size, in words
- `segment_size_window` : the "window" into which looking for the best segment
  split, in words. If `NULL`, it is set to `0.4 * segment_size`.

The result of the function is a `quanteda` corpus, which keeps the
original corpus metadata and adds an additional `segment_source` variable, which keeps track of which segment belongs to which document.

```{r}
corpus
```

```{r}
head(docvars(corpus))
```

### dfm computation

The next step is to compute the document-feature matrix. As our `corpus` object is
a `quanteda` corpus, we can tokenize it and then use the `dfm()` function.

```{r}
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_tolower(tok)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok)
```

We also filter out terms that appear in less than 10 segments by using `dfm_trim`.

```{r}
dtm <- dfm_trim(dtm, min_docfreq = 10)
```

## Simple clustering

We are now ready to compute a simple Reinert clustering by using the
`rainette()` function. Its main arguments are :

- `k` : the number of clusters to compute.
- `min_segment_size` : the minimum number of terms in each segment. 
  If a segment contains less than this number of terms, it will be merged 
  with the following or previous one if they come from the same source document. 
  The default value is 0, *ie* no merging is done.
- `min_split_members` : if a cluster is smaller than this value, it won't be
  split afterwards (default : 5).
  
Here we will compute 5 clusters with a `min_segment_size` of 15 :

```{r message = FALSE, warning = FALSE}
res <- rainette(dtm, k = 5, min_segment_size = 15)
```

To help exploring the clustering results, `rainette` provides an interactive interface which can be launched with `rainette_explor()` :

```{r eval = FALSE}
rainette_explor(res, dtm, corpus)
```

![rainette_explor() interface](img/rainette_explor_plot_en.png)

The interface allows to change the number of clusters, the displayed statistic, etc., and
see the result in real time. By default the most specific terms are displayed with
a blue bar or a red one for those with a negative keyness (if *Show negative values* 
has been checked).

You can also click on the *Get R code* button to get the R code to reproduce the current plot and to compute cluster membership.

The *Cluster documents* tab allows to browse the documents from a given cluster. You can filter them by giving a term or a regular expression in the *Filter by term* field :

![rainette_explor() cluster documents tab](img/rainette_explor_docs_en.png)

You can use `cutree` to get each document cluster at
level `k` :

```{r}
cluster <- cutree(res, k = 5)
```

This vector can be used, for example, as a new corpus metadata variable :

```{r}
corpus$cluster <- cutree(res, k = 5)
head(docvars(corpus))
```

Here, the clusters have been assigned to the segments, not to the original documents as a whole. The `clusters_by_doc_table` allows to display, for each original document, the number of segment belonging to each cluster :

```{r}
clusters_by_doc_table(corpus, clust_var = "cluster")
```

By adding `prop = TRUE`, the same table is displayed with row percentages :

```{r}
clusters_by_doc_table(corpus, clust_var = "cluster", prop = TRUE)
```

Conversely, the `docs_by_cluster_table` allows to display, for each cluster, the number and proportion of original document including at least one segment of this cluster :

```{r}
docs_by_cluster_table(corpus, clust_var = "cluster")
```


## Double clustering

`rainette` also provides a "double clustering" algorithm, as described in the [algorithms description](algorithms_en.html) vignette : two simple clusterings are computed with different `min_segment_size` values, and then crossed together to get more robust clusters.

This can be done with the `rainette2()` function, which can be applied to two already 
computed simple clusterings. Here, we compute them with `min_segment_size` at 10 and 15.

```{r message=FALSE, warning=FALSE}
res1 <- rainette(dtm, k = 5, min_segment_size = 10)
res2 <- rainette(dtm, k = 5, min_segment_size = 15)
```

We then use `rainette2()` to combine them. The `max_k` argument is used to specify the maximum number of clusters.

```{r message=FALSE, warning=FALSE}
res <- rainette2(res1, res2, max_k = 5)
```

One important argument of `rainette2()` is the `full` argument :

- If `full = TRUE` (the default), the best crossed clusters partition selection is made by keeping all non empty crossed clusters. This allows an exhaustive search and to identify the partition either with the highest sum of association χ², or the one with the highest total size. However, computation times can rise rapidly with the number of clusters.
- If `full = FALSE`, only the crossed clusters whose clusters are the most mutually associated are kept. Computations are much faster, but the highest *k* level with available partitions may be lower, and only the partitions with the highest association χ² values can be identified.

To be noted : with `full = TRUE`, if runtime is too high you can add the `parallel = TRUE` argument to paralellise some computations (won't work on Windows, though, and may use much more RAM).

In both cases, the resulting object is a *tibble* with, for each level *k*, the optimal
partitions and their characteristics. Another interactive interface is
available to explore the results. It is launched with `rainette2_explor()`.

```{r eval=FALSE}
rainette2_explor(res, dtm, corpus)
```

![rainette2_explor() interface](img/rainette2_explor_en.png)

The interface is very similar to the previous one, except there is no
dendrogram anymore, but a single barplot of cluster sizes instead. Be
careful of the number of `NA` (not assigned segments), as it can be quite
high. 

If some points are not assigned to any cluster, you can use
`rainette2_complete_groups()` to assign them to the nearest one by using a
*k-nearest-neighbors* algorithm (with k=1). However this may not be recommended as you would then loose the "robustness" of the new clusters computed by `rainette2()`.

```{r}
clusters <- cutree(res, k = 5)
clusters_completed <- rainette2_complete_groups(dtm, clusters)
```

---
title: "[fr] Utilisation de rainette"
author: "Julien Barnier"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{[fr] Utilisation de rainette}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Pour une présentation en français des principes de la méthode et de son utilisation dans `rainette`, vous pouvez aussi consulter la vidéo d'une intervention réalisée dans le cadre d'[un tuto](https://mate-shs.cnrs.fr/actions/tutomate/tuto42-rainette-julien-barnier/) du réseau métier Mate-shs.

<iframe width="560" height="315" src="https://www.youtube.com/embed/T9r8T5WZYHY" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>


# Préparation du corpus

## Importation

La première étape consiste à importer votre corpus dans R. Vous pouvez utiliser un corpus au format `tm` ou `quanteda`, ou bien utiliser la fonction `import_corpus_iramuteq` pour importer directement un corpus depuis un fichier texte au format Iramuteq. Dans ce cas vous obtiendrez un objet de type `corpus` du package [quanteda](https://quanteda.io). Les métadonnées éventuelles présentes dans le fichier (*variables étoilées*) sont importées en tant que métadonnées du corpus (variables de `docvars(corpus)`).

Pour ce document, on va importer le texte du *Manifeste du parti communiste* (Karl Marx et Friedrich Engels, 1848, source [wikisource](https://fr.wikisource.org/wiki/Manifeste_du_parti_communiste/Lafargue)). Celui-ci est placé dans un fichier texte au format Iramuteq fourni avec `rainette`. Le corpus est composé de quatre documents, un document par partie de l'ouvrage.

```{r message=FALSE, warning=FALSE}
library(rainette)
library(quanteda)

## Import du corpus
fichier <- system.file("extdata", "manifeste_pc.txt", package = "rainette")
corpus <- import_corpus_iramuteq(fichier)
```

On peut vérifier que notre corpus est bien constitué de quatre documents (l'introduction et les trois parties principales), et d'une variable de métadonnée :

```{r}
corpus
```

Cette variable de métadonnée est justement la partie correspondant au texte :

```{r paged.print=TRUE}
docvars(corpus)
```


## Découpage en segments

La méthode Reinert de classification s'applique à des segments de texte relativement courts, et non à des textes longs. Une première étape consiste donc à découper chaque texte du corpus en segments via la fonction `split_segments()`. Ici on découpe en segments d'environ 40 mots (l'algorithme essaie de tenir compte de la ponctuation pour, par exemple, placer les césures entre des phrases ou au niveau d'une virgule).

```{r message=FALSE}
corpus <- split_segments(corpus, segment_size = 40)
```

Notre corpus est désormais constitué de 278 segments et 2 variables de métadonnées :

```{r}
corpus
```

Si on regarde les nouvelles métadonnées, on peut voir que la variable `partie` a été dupliquée pour chaque segment correspondant, et une nouvelle variable `segment_source` indique le document d'origine du segment.

```{r}
head(docvars(corpus))
```

On peut aussi visualiser les premiers segments de texte calculés :

```{r}
as.character(corpus)[1:2]
```


## Calcul et traitement de la matrice termes-documents

L'étape suivante est de calculer la matrice termes-documents (*dtm*), grand tableau numérique comportant les documents en lignes, les termes en colonnes, et comme valeurs le nombre d'occurrences de chaque terme dans chaque document.

Notre corpus étant au format `quanteda`, on va utiliser les fonctions de cette extension.

On commence par tokeniser le corpus en supprimant nombre et ponctuations, on supprime les mots outils les plus courants et on convertit tous les termes en minuscules, puis on calule la matrice termes-documents avec la fonction `dfm()`.

```{r}
tok <- tokens(corpus, remove_punct = TRUE, remove_numbers = TRUE)
tok <- tokens_remove(tok, stopwords("fr"))
tok <- tokens_tolower(tok)
dtm <- dfm(tok)
```

On supprime ensuite de la matrice les termes apparaissant dans moins de 3 segments.

```{r}
dtm <- dfm_trim(dtm, min_docfreq = 3)
```

De nombreux autres traitements seraient possibles, mais on se contentera de cette matrice pour cet exemple.


# Classification simple

Une fois notre matrice calculée, on peut procéder à une première forme de classification : une classification descendante hiérarchique simple, calculée avec la fonction `rainette()`.

On peut lui passer plusieurs arguments, notamment :

- `k` : le nombre maximal de classes souhaitées.
- `min_split_members` : le nombre minimal de documents pour qu'une classe soit scindée en deux à l'étape suivante de la classification.
- `min_segment_size` : le nombre minimal de termes par segment. En effet, lors de la tokenisation et du calcul de la dtm, certaines formes (mots-outils, mots trop peu fréquents) ont été supprimées, nos segments peuvent donc varier en taille (entendue comme le nombre de termes encore présents). Avec `min_segment_size = 10`, les segments comportant moins de 10 formes sont regroupés avec le segment suivant ou précédent du même document (si possible) jusqu'à atteindre la taille minimale souhaitée.

```{r message=FALSE}
res <- rainette(dtm, k = 5, min_segment_size = 10, min_split_members = 10)
```

L'objet résultat ne nous dit pas grand chose en lui-même :

```{r}
res
```

Pour faciliter l'exploration des résultats, `rainette` propose une interface interactive qui peut être lancée avec la fonction `rainette_explor()` :

```{r eval = FALSE}
rainette_explor(res, dtm, corpus)
```

L'interface devrait ressembler à quelque chose comme ça :

![Interface de rainette_explor()](img/rainette_explor_plot_fr.png)

Il est possible de modifier le nombre de classes, la statistique utilisée dans les graphiques (spécificité, termes les plus fréquents), etc. Par défaut, les graphiques sous chaque classe indiquent les termes les plus caractéristiques du groupe positivement (en bleu) ou négativement (en rouge et si vous avez coché la case *Show negative values*).

Cette interface vous permet d'expérimenter librement sur le nombre de classes et leur interprétation.

L'onglet *Cluster documents* permet à tout moment de visualiser les documents d'une classe. Vous pouvez également filtrer ces documents en saisissant un mot ou une expression régulière dans le champ *Filter by term* :

![Onglet "Cluster documents" de rainette_explor()](img/rainette_explor_docs_fr.png)

Dans l'onglet *Summary*, vous pouvez également cliquer sur *Get R code* pour obtenir le code R correspondant au graphique actuellement affiché, ainsi que la commande `cutree` qui vous permet de récupérer les groupes d'appartenance de chaque document du corpus, là aussi selon le nombre de groupes actuellement affichés.

```{r eval=FALSE}
## Clustering description plot
rainette_plot(res, dtm, k = 5, type = "bar", n_terms = 20, free_scales = FALSE,
    measure = "chi2", show_negative = "TRUE", text_size = 11)
## Groups
cutree(res, k = 5)
```

Vous pouvez par exemple utiliser l'appel de `cutree` pour ajouter comme nouvelle métadonnée du corpus le groupe d'appartenance de chaque segment pour la classification en 5 classes :

```{r}
corpus$groupe <- cutree(res, k = 5)
head(docvars(corpus))
```

Ici les classes ont attribuées aux segments, et non aux documents dans leur ensemble. La fonction `clusters_by_doc_table()` permet d'afficher, pour chaque document (ici chacune des quatre parties du texte), le nombre de segments de chaque groupe :

```{r}
clusters_by_doc_table(corpus, clust_var = "groupe")
```

En ajoutant `prop = TRUE`, cette répartition peut être visualisée en pourcentages :

```{r}
clusters_by_doc_table(corpus, clust_var = "groupe", prop = TRUE)
```

À l'inverse, `docs_by_cluster_table()` permet de visualiser, pour chaque groupe, le nombre et la proportion de documents d'origine comportant au moins un segment de ce groupe :

```{r}
docs_by_cluster_table(corpus, clust_var = "groupe")
```



# Classification double

Le deuxième type de classification proposé est une classification double : selon la méthode proposée par Max Reinert, on effectue deux classifications simples en faisant varier la taille minimale des segments, puis on "croise" les résultats de ces deux classifications pour déterminer de nouvelles classes, potentiellement plus robustes. Pour plus de détails sur la classifcation double, se référer à la [vignette de description des algorithmes](algorithmes.html).

Une classification double utilise la fonction `rainette2()`. On commence par effectuer les deux classifications simples, ici une avec une taille de segment minimale à 10, et une autre à 15 :

```{r message=FALSE, warning=FALSE}
res1 <- rainette(dtm, k = 7, min_segment_size = 10)
res2 <- rainette(dtm, k = 7, min_segment_size = 15)
```

Puis on utilise `rainette2()` sur ces deux objets résultats, en lui indiquant le nombre maximal de classes à calculer (argument `max_k`).

```{r message=FALSE}
res <- rainette2(res1, res2, max_k = 7)
```

Un paramètre important de `rainette2()` est le paramètre `full` :

- Si `full = TRUE` (valeur par défaut), la sélection des partitions optimales de classes croisées se fait en conservant toutes les classes croisées non nulles. Ceci permet une recherche "exhaustive" et permet de sélectionner les partitions ayant l'effectif total le plus élevé ou bien le χ² d'association le plus fort. Les temps de calcul peuvent cependant être longs si le nombre de classes de départ est important.
- Si `full = FALSE`, seules les classes croisées dont les deux memebres sont les plus associés mutuellement au sens du χ² sont conservées. Les calculs sont beaucoup plus rapides, mais le *k* maximal risque d'être moins élevé, et seules les partitions avec le χ² d'association le plus élevé sont conservées.

À noter que pour `full = TRUE`, si le temps d'exécution est trop long il est possible d'utiliser également `parallel = TRUE` qui permet de paralléliser une partie des calculs (attention cette option ne fonctionne pas sous Windows, et elle utilise davantage de RAM).

Dans tous les cas l'objet résultat contient, pour chaque valeur de `k`, les partitions optimales trouvées et leurs caractéristiques. Là encore, une interface interactive est proposée pour visualiser et explorer ces résultats. Elle se lance via la fonction `rainette2_explor()` :

```{r eval=FALSE}
rainette2_explor(res, dtm, corpus)
```

![Interface de rainette2_explor()](img/rainette2_explor_fr.png)

L'interface est très semblable à la précédente, sauf qu'il n'y a plus de dendrogramme mais à la place un diagramme en barre des effectifs des groupes. Soyez attentifs aux `NA`, qui représentent les segments non classés : contrairement à la classification simple, ils peuvent ici être assez nombreux.

Si certains segments n'ont pas été affecté à un groupe, vous pouvez utiliser `rainette2_complete_groups()` pour les assigner au groupe le plus proche selon une méthode *k-nearest neighbors*  (avec k=1). Ceci n'est cependant pas forcément conseillé car les classes perdont du coup une partie de la "robustesse" acquise par la classification double.

```{r}
groupes <- cutree(res, k = 5)
groupes_complets <- rainette2_complete_groups(dtm, groupes)
```


# Traitements complémentaires

Dans sa version originale, la méthode Reinert de classification est souvent suivie par une analyse complémentaire centrée sur les mots-outils et par différentes représentations graphiques, notamment en projetant les classes obtenues sur des résultats d'analyses factorielles.

On trouvera une explication et un exemple d'application, ainsi qu'un aperçu des fondements théoriques de la méthode dans l'article de Claire Tissot : [Analyse textuelle du roman « Sarrasine » d’Honoré de Balzac : de la recherche de thèmes à la sémiotique](https://clairetissotsemiotique.wordpress.com/2022/05/06/analyse-textuelle-du-roman-sarrasine-dhonore-de-balzac-de-la-recherche-de-themes-a-la-semiotique/).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette.R
\name{rainette}
\alias{rainette}
\title{Corpus clustering based on the Reinert method - Simple clustering}
\usage{
rainette(
  dtm,
  k = 10,
  min_segment_size = 0,
  doc_id = NULL,
  min_split_members = 5,
  cc_test = 0.3,
  tsj = 3,
  min_members,
  min_uc_size
)
}
\arguments{
\item{dtm}{quanteda dfm object of documents to cluster, usually the
result of \code{\link[=split_segments]{split_segments()}}}

\item{k}{maximum number of clusters to compute}

\item{min_segment_size}{minimum number of forms by document}

\item{doc_id}{character name of a dtm docvar which identifies source documents.}

\item{min_split_members}{don't try to split groups with fewer members}

\item{cc_test}{contingency coefficient value for feature selection}

\item{tsj}{minimum frequency value for feature selection}

\item{min_members}{deprecated, use \code{min_split_members} instead}

\item{min_uc_size}{deprecated, use \code{min_segment_size} instead}
}
\value{
The result is a list of both class \code{hclust} and \code{rainette}. Besides the elements
of an \code{hclust} object, two more results are available :
\itemize{
\item \code{uce_groups} give the group of each document for each k
\item \code{group} give the group of each document for the maximum value of k available
}
}
\description{
Corpus clustering based on the Reinert method - Simple clustering
}
\details{
See the references for original articles on the method. Computations and results may differ
quite a bit, see the package vignettes for more details.

The dtm object is automatically converted to boolean.

If \code{min_segment_size > 0} then \code{doc_id} must be provided unless the corpus comes from \code{split_segments},
in this case \code{segment_source} is used by default.
}
\examples{
\donttest{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 3)
res <- rainette(dtm, k = 3, min_segment_size = 15)
}

}
\references{
\itemize{
\item Reinert M, Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte, Cahiers de l'analyse des données, Volume 8, Numéro 2, 1983. \url{https://www.numdam.org/item/?id=CAD_1983__8_2_187_0}
\item Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Méthodologie Sociologique, Volume 26, Numéro 1, 1990. \doi{10.1177/075910639002600103}
}
}
\seealso{
\code{\link[=split_segments]{split_segments()}}, \code{\link[=rainette2]{rainette2()}}, \code{\link[=cutree_rainette]{cutree_rainette()}}, \code{\link[=rainette_plot]{rainette_plot()}}, \code{\link[=rainette_explor]{rainette_explor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats.R
\name{rainette_stats}
\alias{rainette_stats}
\title{Generate cluster keyness statistics from a rainette result}
\usage{
rainette_stats(
  groups,
  dtm,
  measure = c("chi2", "lr", "frequency", "docprop"),
  n_terms = 15,
  show_negative = TRUE,
  max_p = 0.05
)
}
\arguments{
\item{groups}{groups membership computed by \code{cutree_rainette} or \code{cutree_rainette2}}

\item{dtm}{the dfm object used to compute the clustering}

\item{measure}{statistics to compute}

\item{n_terms}{number of terms to display in keyness plots}

\item{show_negative}{if TRUE, show negative keyness features}

\item{max_p}{maximum keyness statistic p-value}
}
\value{
A list with, for each group, a data.frame of keyness statistics for the most specific
n_terms features.
}
\description{
Generate cluster keyness statistics from a rainette result
}
\examples{
\donttest{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 3)
res <- rainette(dtm, k = 3, min_segment_size = 15)
groups <- cutree_rainette(res, k = 3)
rainette_stats(groups, dtm)
}
}
\seealso{
\code{\link[quanteda.textstats:textstat_keyness]{quanteda.textstats::textstat_keyness()}}, \code{\link[=rainette_explor]{rainette_explor()}}, \code{\link[=rainette_plot]{rainette_plot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette.R
\name{order_docs}
\alias{order_docs}
\title{return documents indices ordered by CA first axis coordinates}
\usage{
order_docs(m)
}
\arguments{
\item{m}{dtm on which to compute the CA and order documents, converted to an integer matrix.}
}
\value{
ordered list of document indices
}
\description{
return documents indices ordered by CA first axis coordinates
}
\details{
Internal function, not to be used directly
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{rainette_plot}
\alias{rainette_plot}
\title{Generate a clustering description plot from a rainette result}
\usage{
rainette_plot(
  res,
  dtm,
  k = NULL,
  type = c("bar", "cloud"),
  n_terms = 15,
  free_scales = FALSE,
  measure = c("chi2", "lr", "frequency", "docprop"),
  show_negative = FALSE,
  text_size = NULL,
  show_na_title = TRUE,
  cluster_label = NULL,
  keyness_plot_xlab = NULL,
  colors = NULL
)
}
\arguments{
\item{res}{result object of a \code{rainette} clustering}

\item{dtm}{the dfm object used to compute the clustering}

\item{k}{number of groups. If NULL, use the biggest number possible}

\item{type}{type of term plots : barplot or wordcloud}

\item{n_terms}{number of terms to display in keyness plots}

\item{free_scales}{if TRUE, all the keyness plots will have the same scale}

\item{measure}{statistics to compute}

\item{show_negative}{if TRUE, show negative keyness features}

\item{text_size}{font size for barplots, max word size for wordclouds}

\item{show_na_title}{if TRUE, show number of NA as plot title}

\item{cluster_label}{define a specific term for clusters identification in keyness plots.
Default is "Cluster" or "Cl." depending on the number of groups. If a vector of length > 1,
define the cluster labels manually.}

\item{keyness_plot_xlab}{define a specific x label for keyness plots.}

\item{colors}{vector of custom colors for cluster titles and branches (in the order of the clusters)}
}
\value{
A gtable object.
}
\description{
Generate a clustering description plot from a rainette result
}
\examples{
\donttest{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 3)
res <- rainette(dtm, k = 3, min_segment_size = 15)
rainette_plot(res, dtm)
rainette_plot(
  res,
  dtm,
  cluster_label = c("Assets", "Future", "Values"),
  colors = c("red", "slateblue", "forestgreen")
)
}

}
\seealso{
\code{\link[quanteda.textstats:textstat_keyness]{quanteda.textstats::textstat_keyness()}}, \code{\link[=rainette_explor]{rainette_explor()}}, \code{\link[=rainette_stats]{rainette_stats()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cutree.R
\name{cutree}
\alias{cutree}
\title{Cut a tree into groups}
\usage{
cutree(tree, ...)
}
\arguments{
\item{tree}{the hclust tree object to be cut}

\item{...}{arguments passed to other methods}
}
\value{
A vector with group membership.
}
\description{
Cut a tree into groups
}
\details{
If \code{tree} is of class \code{rainette}, invokes \code{\link[=cutree_rainette]{cutree_rainette()}}. Otherwise, just run \code{\link[stats:cutree]{stats::cutree()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette2_explor.R
\name{rainette2_explor}
\alias{rainette2_explor}
\title{Shiny gadget for rainette2 clustering exploration}
\usage{
rainette2_explor(res, dtm = NULL, corpus_src = NULL)
}
\arguments{
\item{res}{result object of a \code{rainette2} clustering}

\item{dtm}{the dfm object used to compute the clustering}

\item{corpus_src}{the quanteda corpus object used to compute the dtm}
}
\value{
No return value, called for side effects.
}
\description{
Shiny gadget for rainette2 clustering exploration
}
\seealso{
\code{\link[=rainette2_plot]{rainette2_plot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette_explor.R
\name{rainette_explor}
\alias{rainette_explor}
\title{Shiny gadget for rainette clustering exploration}
\usage{
rainette_explor(res, dtm = NULL, corpus_src = NULL)
}
\arguments{
\item{res}{result object of a \code{rainette} clustering}

\item{dtm}{the dfm object used to compute the clustering}

\item{corpus_src}{the quanteda corpus object used to compute the dtm}
}
\value{
No return value, called for side effects.
}
\description{
Shiny gadget for rainette clustering exploration
}
\examples{
\dontrun{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 3)
res <- rainette(dtm, k = 3, min_segment_size = 15)
rainette_explor(res, dtm, corpus)
}

}
\seealso{
\code{rainette_plot}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette.R
\name{select_features}
\alias{select_features}
\title{Remove features from dtm of each group base don cc_test and tsj}
\usage{
select_features(m, indices1, indices2, cc_test = 0.3, tsj = 3)
}
\arguments{
\item{m}{global dtm}

\item{indices1}{indices of documents of group 1}

\item{indices2}{indices of documents of group 2}

\item{cc_test}{maximum contingency coefficient value for the
feature to be kept in both groups.}

\item{tsj}{minimum feature frequency in the dtm}
}
\value{
a list of two character vectors : \code{cols1} is the name of features to
keep in group 1, \code{cols2} the name of features to keep in group 2
}
\description{
Remove features from dtm of each group base don cc_test and tsj
}
\details{
Internal function, not to be used directly
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette.R
\name{cluster_tab}
\alias{cluster_tab}
\title{Split a dtm into two clusters with reinert algorithm}
\usage{
cluster_tab(dtm, cc_test = 0.3, tsj = 3)
}
\arguments{
\item{dtm}{to be split, passed by \code{rainette}}

\item{cc_test}{maximum contingency coefficient value for the
feature to be kept in both groups.}

\item{tsj}{minimum feature frequency in the dtm}
}
\value{
An object of class \code{hclust} and \code{rainette}
}
\description{
Split a dtm into two clusters with reinert algorithm
}
\details{
Internal function, not to be used directly
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette.R
\name{switch_docs}
\alias{switch_docs}
\title{Switch documents between two groups to maximize chi-square value}
\usage{
switch_docs(m, indices, max_index, max_chisq)
}
\arguments{
\item{m}{original dtm}

\item{indices}{documents indices orderes by first CA axis coordinates}

\item{max_index}{document index where the split is maximum}

\item{max_chisq}{maximum chi-square value}
}
\value{
a list of two vectors \code{indices1} and \code{indices2}, which contain
the documents indices of each group after documents switching, and a \code{chisq} value,
the new corresponding chi-square value after switching
}
\description{
Switch documents between two groups to maximize chi-square value
}
\details{
Internal function, not to be used directly
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{clusters_by_doc_table}
\alias{clusters_by_doc_table}
\title{Returns the number of segment of each cluster for each source document}
\usage{
clusters_by_doc_table(obj, clust_var = NULL, doc_id = NULL, prop = FALSE)
}
\arguments{
\item{obj}{a corpus, tokens or dtm object}

\item{clust_var}{name of the docvar with the clusters}

\item{doc_id}{docvar identifying the source document}

\item{prop}{if TRUE, returns the percentage of each cluster by document}
}
\description{
Returns the number of segment of each cluster for each source document
}
\details{
This function is only useful for previously segmented corpus. If \code{doc_id}
is NULL and there is a \code{sement_source} docvar, it will be used instead.
}
\examples{
\donttest{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 2)
res <- rainette(dtm, k = 3, min_segment_size = 15)
corpus$cluster <- cutree(res, k = 3)
clusters_by_doc_table(corpus, clust_var = "cluster", prop = TRUE)
}
}
\seealso{
\code{\link[=docs_by_cluster_table]{docs_by_cluster_table()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{rainette2_plot}
\alias{rainette2_plot}
\title{Generate a clustering description plot from a rainette2 result}
\usage{
rainette2_plot(
  res,
  dtm,
  k = NULL,
  criterion = c("chi2", "n"),
  complete_groups = FALSE,
  type = c("bar", "cloud"),
  n_terms = 15,
  free_scales = FALSE,
  measure = c("chi2", "lr", "frequency", "docprop"),
  show_negative = FALSE,
  text_size = 10
)
}
\arguments{
\item{res}{result object of a \code{rainette2} clustering}

\item{dtm}{the dfm object used to compute the clustering}

\item{k}{number of groups. If NULL, use the biggest number possible}

\item{criterion}{criterion to use to choose the best partition. \code{chi2} means
the partition with the maximum sum of chi2, \code{n} the partition with the
maximum size.}

\item{complete_groups}{if TRUE, documents with NA cluster are reaffected by
k-means clustering initialised with current groups centers.}

\item{type}{type of term plots : barplot or wordcloud}

\item{n_terms}{number of terms to display in keyness plots}

\item{free_scales}{if TRUE, all the keyness plots will have the same scale}

\item{measure}{statistics to compute}

\item{show_negative}{if TRUE, show negative keyness features}

\item{text_size}{font size for barplots, max word size for wordclouds}
}
\value{
A gtable object.
}
\description{
Generate a clustering description plot from a rainette2 result
}
\seealso{
\code{\link[quanteda.textstats:textstat_keyness]{quanteda.textstats::textstat_keyness()}}, \code{\link[=rainette2_explor]{rainette2_explor()}}, \code{\link[=rainette2_complete_groups]{rainette2_complete_groups()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_segments.R
\name{split_segments}
\alias{split_segments}
\alias{split_segments.character}
\alias{split_segments.Corpus}
\alias{split_segments.corpus}
\alias{split_segments.tokens}
\title{Split a character string or corpus into segments}
\usage{
split_segments(obj, segment_size = 40, segment_size_window = NULL)

\method{split_segments}{character}(obj, segment_size = 40, segment_size_window = NULL)

\method{split_segments}{Corpus}(obj, segment_size = 40, segment_size_window = NULL)

\method{split_segments}{corpus}(obj, segment_size = 40, segment_size_window = NULL)

\method{split_segments}{tokens}(obj, segment_size = 40, segment_size_window = NULL)
}
\arguments{
\item{obj}{character string, quanteda or tm corpus object}

\item{segment_size}{segment size (in words)}

\item{segment_size_window}{window around segment size to look for best splitting point}
}
\value{
If obj is a tm or quanteda corpus object, the result is a quanteda corpus.
}
\description{
Split a character string or corpus into segments, taking into account punctuation where possible
}
\examples{
\donttest{
require(quanteda)
split_segments(data_corpus_inaugural)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cutree.R
\name{cutree_rainette2}
\alias{cutree_rainette2}
\title{Cut a rainette2 result object into groups of documents}
\usage{
cutree_rainette2(res, k, criterion = c("chi2", "n"), ...)
}
\arguments{
\item{res}{the \code{rainette2} result object to be cut}

\item{k}{the desired number of clusters}

\item{criterion}{criterion to use to choose the best partition. \code{chi2} means
the partition with the maximum sum of chi2, \code{n} the partition with the
maximum size.}

\item{...}{arguments passed to other methods}
}
\value{
A vector with group membership.
}
\description{
Cut a rainette2 result object into groups of documents
}
\seealso{
\code{\link[=rainette2_complete_groups]{rainette2_complete_groups()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{merge_segments}
\alias{merge_segments}
\title{Merges segments according to minimum segment size}
\usage{
merge_segments(dtm, min_segment_size = 10, doc_id = NULL)
}
\arguments{
\item{dtm}{dtm of segments}

\item{min_segment_size}{minimum number of forms by segment}

\item{doc_id}{character name of a dtm docvar which identifies source documents.}
}
\value{
the original dtm with a new \code{rainette_uc_id} docvar.
}
\description{
\code{rainette_uc_index} docvar
}
\details{
If \code{min_segment_size == 0}, no segments are merged together.
If \code{min_segment_size > 0} then \code{doc_id} must be provided
unless the corpus comes from \code{split_segments}, in this case
\code{segment_source} is used by default.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rainette2.R
\name{rainette2}
\alias{rainette2}
\title{Corpus clustering based on the Reinert method - Double clustering}
\usage{
rainette2(
  x,
  y = NULL,
  max_k = 5,
  min_segment_size1 = 10,
  min_segment_size2 = 15,
  doc_id = NULL,
  min_members = 10,
  min_chi2 = 3.84,
  parallel = FALSE,
  full = TRUE,
  uc_size1,
  uc_size2,
  ...
)
}
\arguments{
\item{x}{either a quanteda dfm object or the result of \code{\link[=rainette]{rainette()}}}

\item{y}{if \code{x} is a \code{\link[=rainette]{rainette()}} result, this must be another \code{\link[=rainette]{rainette()}}
result from same dfm but with different uc size.}

\item{max_k}{maximum number of clusters to compute}

\item{min_segment_size1}{if \code{x} is a dfm, minimum uc size for first clustering}

\item{min_segment_size2}{if \code{x} is a dfm, minimum uc size for second clustering}

\item{doc_id}{character name of a dtm docvar which identifies source documents.}

\item{min_members}{minimum members of each cluster}

\item{min_chi2}{minimum chi2 for each cluster}

\item{parallel}{if TRUE, use \code{parallel::mclapply} to compute partitions
(won't work on Windows, uses more RAM)}

\item{full}{if TRUE, all crossed groups are kept to compute optimal partitions, otherwise
only the most mutually associated groups are kept.}

\item{uc_size1}{deprecated, use min_segment_size1 instead}

\item{uc_size2}{deprecated, use min_segment_size2 instead}

\item{...}{if \code{x} is a dfm object, parameters passed to \code{\link[=rainette]{rainette()}} for both
simple clusterings}
}
\value{
A tibble with optimal partitions found for each available value of \code{k} as rows, and the following
columns :
\itemize{
\item \code{clusters} list of the crossed original clusters used in the partition
\item \code{k} the number of clusters
\item \code{chi2} sum of the chi2 value of each cluster
\item \code{n} sum of the size of each cluster
\item \code{groups} group membership of each document for this partition (\code{NA} if not assigned)
}
}
\description{
Corpus clustering based on the Reinert method - Double clustering
}
\details{
You can pass a quanteda dfm as \code{x} object, the function then performs two simple
clustering with varying minimum uc size, and then proceed to find optimal partitions
based on the results of both clusterings.

If both clusterings have already been computed, you can pass them as \code{x} and \code{y} arguments
and the function will only look for optimal partitions.

\code{doc_id} must be provided unless the corpus comes from \code{split_segments},
in this case \code{segment_source} is used by default.

If \code{full = FALSE}, computation may be much faster, but the chi2 criterion will be the only
one available for best partition detection, and the result may not be optimal.

For more details on optimal partitions search algorithm, please see package vignettes.
}
\examples{
\donttest{
require(quanteda)
corpus <- data_corpus_inaugural
corpus <- head(corpus, n = 10)
corpus <- split_segments(corpus)
tok <- tokens(corpus, remove_punct = TRUE)
tok <- tokens_remove(tok, stopwords("en"))
dtm <- dfm(tok, tolower = TRUE)
dtm <- dfm_trim(dtm, min_docfreq = 3)

res1 <- rainette(dtm, k = 5, min_segment_size = 10)
res2 <- rainette(dtm, k = 5, min_segment_size = 15)

res <- rainette2(res1, res2, max_k = 4)
}

}
\references{
\itemize{
\item Reinert M, Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte, Cahiers de l'analyse des données, Volume 8, Numéro 2, 1983. \url{https://www.numdam.org/item/?id=CAD_1983__8_2_187_0}
\item Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Méthodologie Sociologique, Volume 26, Numéro 1, 1990.  \doi{10.1177/075910639002600103}
}
}
\seealso{
\code{\link[=rainette]{rainette()}}, \code{\link[=cutree_rainette2]{cutree_rainette2()}}, \code{\link[=rainette2_plot]{rainette2_plot()}}, \code{\link[=rainette2_explor]{rainette2_explor()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cutree.R
\name{rainette2_complete_groups}
\alias{rainette2_complete_groups}
\title{Complete groups membership with knn classification}
\usage{
rainette2_complete_groups(dfm, groups, k = 1, ...)
}
\arguments{
\item{dfm}{dfm object used for \code{rainette2} clustering.}

\item{groups}{group membership computed by \code{cutree} on \code{rainette2} result.}

\item{k}{number of neighbours considered.}

\item{...}{other arguments passed to \code{FNN::knn}.}
}
\value{
Completed group membership vector.
}
\description{
Starting with groups membership computed from a \code{rainette2} clustering,
every document not assigned to a cluster is reassigned using a k-nearest
neighbour classification.
}
\seealso{
\code{\link[=cutree_rainette2]{cutree_rainette2()}}, \code{\link[FNN:knn]{FNN::knn()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import.R
\name{import_corpus_iramuteq}
\alias{import_corpus_iramuteq}
\title{Import a corpus in Iramuteq format}
\usage{
import_corpus_iramuteq(f, id_var = NULL, thematics = c("remove", "split"), ...)
}
\arguments{
\item{f}{a file name or a connection}

\item{id_var}{name of metadata variable to be used as documents id}

\item{thematics}{if "remove", thematics lines are removed. If "split", texts as splitted at each thematic, and metadata duplicated accordingly}

\item{...}{arguments passed to \code{\link[base:connections]{file}} if \code{f} is a file name.}
}
\value{
A quanteda corpus object. Note that metadata variables in docvars are all imported as characters.
}
\description{
Import a corpus in Iramuteq format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cutree.R
\name{cutree_rainette}
\alias{cutree_rainette}
\title{Cut a rainette result tree into groups of documents}
\usage{
cutree_rainette(hres, k = NULL, h = NULL, ...)
}
\arguments{
\item{hres}{the \code{rainette} result object to be cut}

\item{k}{the desired number of clusters}

\item{h}{unsupported}

\item{...}{arguments passed to other methods}
}
\value{
A vector with group membership.
}
\description{
Cut a rainette result tree into groups of documents
}
